# Unit test _create_predict_inputs ForecasterRecursiveMultiSeries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.compose import make_column_transformer
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import HistGradientBoostingRegressor
from lightgbm import LGBMRegressor

from ....recursive import ForecasterRecursiveMultiSeries

# Fixtures
from .fixtures_forecaster_recursive_multiseries import (
    series_wide_dt,
    series_long_dt,
    series_dict_range,
    series_dict_dt,
    exog_wide_range,
    exog_long_dt,
    exog_pred_wide_range,
    exog_pred_long_dt,
    series_dict_nans_train,
    exog_dict_nans_train,
    exog_dict_nans_test
)


def test_create_predict_inputs_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster._create_predict_inputs(steps=5)


def test_create_predict_inputs_TypeError_when_exog_MultiIndex_no_DatetimeIndex():
    """
    Test NotFittedError is raised when fitted is False.
    """
    exog_long_no_index = exog_long_dt.copy()
    exog_long_no_index = exog_long_no_index.reset_index(level=1, drop=True)
    exog_long_no_index['idx'] = pd.RangeIndex(
        start=0, stop=len(exog_long_no_index), step=1
    )
    exog_long_no_index = exog_long_no_index.set_index("idx", append=True)

    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)
    forecaster.fit(series=series_long_dt, exog=exog_long_dt)

    err_msg = re.escape(
        f"When `exog` is a pandas MultiIndex DataFrame, its index "
        f"must be a pandas DatetimeIndex. If you want to use a pandas "
        f"RangeIndex, use a dictionary instead. Found `exog` index "
        f"type: {type(exog_long_no_index.index.levels[1])}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster._create_predict_inputs(steps=5, exog=exog_long_no_index)


def test_output_create_predict_inputs_when_regressor_is_LinearRegression():
    """
    Test output _create_predict_inputs when using LinearRegression as regressor.
    """
    series = {
        '1': pd.Series(np.arange(start=0, stop=50, dtype=float)), 
        '2': pd.Series(np.arange(start=50, stop=100, dtype=float))
    }
    
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=5, transformer_series=None
    )
    forecaster.fit(series=series)
    results = forecaster._create_predict_inputs(steps=5)

    expected = (
        pd.DataFrame(
            {'1': np.array([45., 46., 47., 48., 49.]),
             '2': np.array([95., 96., 97., 98., 99.])},
            index = pd.RangeIndex(start=45, stop=50, step=1)
        ),
        None,
        ['1', '2'],
        pd.RangeIndex(start=50, stop=55, step=1)
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_output_when_regressor_is_LinearRegression_with_transform_series():
    """
    Test _create_predict_inputs output when using LinearRegression as regressor 
    and StandardScaler.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor          = LinearRegression(),
                     lags               = 5,
                     transformer_series = StandardScaler()
                 )
    forecaster.fit(series=series_long_dt)
    results = forecaster._create_predict_inputs(steps=5, levels='1')
    
    expected = (
        pd.DataFrame(
            {'1': np.array([
                      -1.0814183452563133, -0.08097053361357116, 2.0806640273868244, 
                       0.07582436346198025, 0.4776288428854555])},
            index = pd.date_range(start='2000-02-15', periods=5, freq='D')
        ),
        None,
        ['1'],
        pd.date_range(start='2000-02-20', periods=5, freq='D')
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


@pytest.mark.parametrize("transformer_series", 
                         [StandardScaler(),
                          {'l1': StandardScaler(), 'l2': StandardScaler(), '_unknown_level': StandardScaler()}], 
                         ids = lambda tr: f'transformer_series type: {type(tr)}')
def test_create_predict_inputs_output_when_regressor_is_LinearRegression_with_transform_series_and_transform_exog(transformer_series):
    """
    Test _create_predict_inputs output when using LinearRegression as regressor, 
    StandardScaler as transformer_series and transformer_exog as transformer_exog.
    """
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor          = LinearRegression(),
                     lags               = 5,
                     transformer_series = transformer_series,
                     transformer_exog   = transformer_exog,
                 )
    forecaster.fit(series=series_dict_dt, exog=exog_long_dt)
    results = forecaster._create_predict_inputs(
        steps=5, levels='l1', exog=exog_pred_long_dt
    )
    
    expected = (
        pd.DataFrame(
            {'l1': np.array([
                      -1.0814183452563133, -0.08097053361357116, 2.0806640273868244, 
                       0.07582436346198025, 0.4776288428854555])},
            index = pd.date_range(start='2000-02-15', periods=5, freq='D')
        ),
        {1: np.array([[-0.09362908,  1., 0.]]),
         2: np.array([[0.45144522,  1., 0.]]),
         3: np.array([[-1.53968887,  1., 0.]]),
         4: np.array([[-1.45096055,  1., 0.]]),
         5: np.array([[-0.77240468,  1., 0.]])
        },
        ['l1'],
        pd.date_range(start='2000-02-20', periods=5, freq='D')
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    for k in results[1].keys():
        np.testing.assert_array_almost_equal(results[1][k], expected[1][k])
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


@pytest.mark.parametrize("transformer_series", 
                         [StandardScaler(),
                          {'l1': StandardScaler(), 'l2': StandardScaler(), '_unknown_level': StandardScaler()}], 
                         ids = lambda tr: f'transformer_series type: {type(tr)}')
def test_create_predict_inputs_output_when_regressor_is_LinearRegression_with_transform_series_and_transform_exog_different_length_series(transformer_series):
    """
    Test _create_predict_inputs output when using LinearRegression as regressor, StandardScaler
    as transformer_series and transformer_exog as transformer_exog with series 
    of different lengths.
    """
    series_dict_range_nan = {
        'l1': series_dict_range['l1'].copy(),
        'l2': series_dict_range['l2'].copy()
    }
    series_dict_range_nan['l2'].iloc[:10] = np.nan

    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor          = LinearRegression(),
                     lags               = 5,
                     transformer_series = transformer_series,
                     transformer_exog   = transformer_exog,
                 )
    forecaster.fit(series=series_dict_range_nan, exog=exog_wide_range)
    results = forecaster._create_predict_inputs(steps=5, exog=exog_pred_wide_range)
    
    expected = (
        pd.DataFrame(
            {'l1': np.array([-1.08141835, -0.08097053, 2.08066403, 0.07582436, 0.47762884]),
             'l2': np.array([0.73973379, -1.29936753, -0.36483376, -0.96090271, -0.57388468])},
            index = pd.RangeIndex(start=45, stop=50, step=1)
        ),
        {1: np.array([[-0.09575486,  1., 0.],
                      [-0.09575486,  1., 0.]]),
         2: np.array([[0.44024531,  1., 0.],
                      [0.44024531,  1., 0.]]),
         3: np.array([[-1.51774136,  1., 0.],
                      [-1.51774136,  1., 0.]]),
         4: np.array([[-1.43049014,  1., 0.],
                      [-1.43049014,  1., 0.]]),
         5: np.array([[-0.76323054,  1., 0.],
                      [-0.76323054,  1., 0.]])
        },
        ['l1', 'l2'],
        pd.RangeIndex(start=50, stop=55, step=1)
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    for k in results[1].keys():
        np.testing.assert_array_almost_equal(results[1][k], expected[1][k])
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_output_when_categorical_features_native_implementation_HistGradientBoostingRegressor():
    """
    Test _create_predict_inputs output when using HistGradientBoostingRegressor and categorical variables.
    """
    df_exog = pd.DataFrame(
        {'exog_1': exog_wide_range['exog_1'],
         'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
         'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)}
    )
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor          = HistGradientBoostingRegressor(
                                              categorical_features = categorical_features,
                                              random_state         = 123
                                          ),
                     lags               = 5,
                     transformer_series = None,
                     transformer_exog   = transformer_exog
                 )
    forecaster.fit(series=series_dict_range, exog=df_exog)
    results = forecaster._create_predict_inputs(steps=10, exog=exog_predict)
    
    expected = (
        pd.DataFrame(
            {'l1': np.array([0.25045537, 0.48303426, 0.98555979, 0.51948512, 0.61289453]),
             'l2': np.array([0.6917018 , 0.15112745, 0.39887629, 0.2408559 , 0.34345601])},
            index = pd.RangeIndex(start=45, stop=50, step=1)
        ),
        {1: np.array([[0, 0, 0.51312815],
                      [0, 0, 0.51312815]]),
         2: np.array([[1, 1, 0.66662455],
                      [1, 1, 0.66662455]]),
         3: np.array([[2, 2, 0.10590849],
                      [2, 2, 0.10590849]]),
         4: np.array([[3, 3, 0.13089495],
                      [3, 3, 0.13089495]]),
         5: np.array([[4, 4, 0.32198061],
                      [4, 4, 0.32198061]]),
         6: np.array([[0, 0, 0.66156434],
                      [0, 0, 0.66156434]]),
         7: np.array([[1, 1, 0.84650623],
                      [1, 1, 0.84650623]]),
         8: np.array([[2, 2, 0.55325734],
                      [2, 2, 0.55325734]]),
         9: np.array([[3, 3, 0.85445249],
                      [3, 3, 0.85445249]]),
         10: np.array([[4, 4, 0.38483781],
                       [4, 4, 0.38483781]])

        },
        ['l1', 'l2'],
        pd.RangeIndex(start=50, stop=60, step=1)
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    for k in results[1].keys():
        np.testing.assert_array_almost_equal(results[1][k], expected[1][k])
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_output_when_series_and_exog_dict():
    """
    Test output ForecasterRecursiveMultiSeries _create_predict_inputs method when 
    series and exog are dictionaries.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        regressor          = LGBMRegressor(
            n_estimators=2, random_state=123, verbose=-1, max_depth=2
        ),
        lags               = 5,
        encoding           = 'ordinal',
        dropna_from_series = False,
        transformer_series = StandardScaler(),
        transformer_exog   = StandardScaler(),
    )
    forecaster.fit(
        series=series_dict_nans_train, exog=exog_dict_nans_train, suppress_warnings=True
    )
    results = forecaster._create_predict_inputs(steps=5, exog=exog_dict_nans_test)
    
    expected = (
        pd.DataFrame(
            {'id_1000': np.array([-0.3525861, -0.457091, -0.49618465, -1.07810218, -1.77580056]),
             'id_1001': np.array([0.21800529, 0.36936645, 0.67896814, 1.16332175, 1.1523137]),
             'id_1003': np.array([-0.62651976, -0.74685959, -1.03823091, -0.37837738, 3.39980134]),
             'id_1004': np.array([0.61495753, 0.8322971, 0.6719899, -0.24309812, -0.75073762])},
            index = pd.date_range(start='2016-07-27', periods=5, freq='D')
        ),
        {1: np.array([[0.00821644, 1.42962482, np.nan, np.nan],
                      [0.00821644, 1.42962482, 1.11141113, -0.87943526],
                      [0.00821644, np.nan, 1.11141113, -0.87943526],
                      [0.00821644, 1.42962482, 1.11141113, -0.87943526]]),
         2: np.array([[1.11220226, 0.89634375, np.nan, np.nan],
                      [1.11220226, 0.89634375, 1.1327558 , 0.0058948 ],
                      [1.11220226, np.nan, 1.1327558 , 0.0058948 ],
                      [1.11220226, 0.89634375, 1.1327558 , 0.0058948 ]]),
         3: np.array([[1.38486425, -0.30192795, np.nan, np.nan],
                      [1.38486425, -0.30192795, 1.1775869 , -0.3532584 ],
                      [1.38486425, np.nan, 1.1775869 , -0.3532584 ],
                      [1.38486425, -0.30192795, 1.1775869 , -0.3532584 ]]),
         4: np.array([[0.62088235, -1.26286725, np.nan, np.nan],
                      [0.62088235, -1.26286725, 1.0428337 , 0.84287284],
                      [0.62088235, np.nan, 1.0428337 , 0.84287284],
                      [0.62088235, -1.26286725, 1.0428337 , 0.84287284]]),
         5: np.array([[-0.60444947, -1.26286725, np.nan, np.nan],
                      [-0.60444947, -1.26286725, 1.00599776, -0.62314633],
                      [-0.60444947, np.nan, 1.00599776, -0.62314633],
                      [-0.60444947, -1.26286725, 1.00599776, -0.62314633]])
        },
        ['id_1000', 'id_1001', 'id_1003', 'id_1004'],
        pd.date_range(start='2016-08-01', periods=5, freq='D')
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    for k in results[1].keys():
        np.testing.assert_array_almost_equal(results[1][k], expected[1][k])
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


@pytest.mark.parametrize("differentiation", 
                         [1, {'1': 1, '2': 1, '_unknown_level': 1}], 
                         ids = lambda diff: f'differentiation: {diff}')
def test_create_predict_inputs_output_when_regressor_is_LinearRegression_with_exog_differentiation_is_1_and_transformer_series(differentiation):
    """
    Test _create_predict_inputs output when using LinearRegression as regressor and differentiation=1,
    and transformer_series is StandardScaler.
    """
    end_train = '2003-01-30 23:59:00'

    # Data scaled and differentiated
    series_datetime = series_wide_dt.copy()
    series_datetime.index = pd.date_range(
        start='2003-01-01', periods=len(series_wide_dt), freq='D'
    )
    series_dict_datetime = {
        "1": series_datetime['1'].loc[:end_train],
        "2": series_datetime['2'].loc[:end_train]
    }
    
    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(series_wide_dt)), name='exog'
    )
    exog.index = pd.date_range(start='2003-01-01', periods=len(series_wide_dt), freq='D')
    exog_dict_datetime = {
        "1": exog.loc[:end_train],
        "2": exog.loc[:end_train]
    }
    exog_pred = {
        '1': exog.loc[end_train:],
        '2': exog.loc[end_train:]
    }

    steps = len(series_datetime.loc[end_train:])

    forecaster = ForecasterRecursiveMultiSeries(
                     regressor          = LinearRegression(), 
                     lags               = 15, 
                     transformer_series = StandardScaler(),    
                     differentiation    = differentiation
                 )
    forecaster.fit(
        series=series_dict_datetime, exog=exog_dict_datetime, store_in_sample_residuals=True
    )
    results = forecaster._create_predict_inputs(
        steps=steps, exog=exog_pred, predict_probabilistic=True
    )
    
    expected = (
        pd.DataFrame(
            {'1': np.array([
                      np.nan, 1.53099926e+00, -2.50175863e+00, -3.17051107e-02,
                      1.60372524e+00, 1.24398054e-03, 4.61947999e-01, 9.68409847e-01,
                      -5.62842287e-01, -5.10849917e-01, 5.01788973e-01, -1.79911269e+00,
                      1.74873121e-01, -6.01343264e-01, 2.94763400e-01, 1.51888875e+00]
                  ),
             '2': np.array([
                      np.nan, -2.93045496, 2.6268138, -2.00736345, -0.19086106,
                      1.46005589, -1.8403251, 3.04806289, -0.99623011, 0.3712246,
                      -2.73026424, 2.23235938, -0.14532345, -1.53568303, -0.02273313,
                      2.09399597]
                  )},
            index = pd.date_range(start='2003-01-15', periods=16, freq='D')
        ),
        {1: np.array([[-0.78916779],
                      [-0.78916779]]),
         2: np.array([[0.07769117],
                      [0.07769117]]),
         3: np.array([[0.27283009],
                      [0.27283009]]),
         4: np.array([[0.35075195],
                      [0.35075195]]),
         5: np.array([[0.31597336],
                      [0.31597336]]),
         6: np.array([[0.14091339],
                      [0.14091339]]),
         7: np.array([[1.28618144],
                      [1.28618144]]),
         8: np.array([[0.09164865],
                      [0.09164865]]),
         9: np.array([[-0.50744682],
                      [-0.50744682]]),
         10: np.array([[0.01522573],
                       [0.01522573]]),
         11: np.array([[0.82813767],
                       [0.82813767]]),
         12: np.array([[0.15290495],
                       [0.15290495]]),
         13: np.array([[0.61532804],
                       [0.61532804]]),
         14: np.array([[-0.06287136],
                       [-0.06287136]]),
         15: np.array([[-1.1896156],
                       [-1.1896156]]),
         16: np.array([[0.28674823],
                       [0.28674823]]),
         17: np.array([[-0.64581528],
                       [-0.64581528]]),
         18: np.array([[0.20879998],
                       [0.20879998]]),
         19: np.array([[1.80029302],
                       [1.80029302]]),
         20: np.array([[0.14269745],
                       [0.14269745]])
        },
        ['1', '2'],
        pd.date_range(start='2003-01-31', periods=steps, freq='D')
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    for k in results[1].keys():
        np.testing.assert_array_almost_equal(results[1][k], expected[1][k])
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


@pytest.mark.parametrize("levels", 
                         [None, ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005']], 
                         ids = lambda levels: f'levels: {levels}')
def test_create_predict_inputs_output_when_series_and_exog_dict_unknown_level(levels):
    """
    Test output ForecasterRecursiveMultiSeries _create_predict_inputs method when 
    series and exog are dictionaries and unknown level.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        regressor          = LGBMRegressor(
            n_estimators=2, random_state=123, verbose=-1, max_depth=2
        ),
        lags               = 5,
        encoding           = 'ordinal',
        dropna_from_series = False,
        transformer_series = StandardScaler(),
        transformer_exog   = StandardScaler(),
    )
    forecaster.fit(
        series=series_dict_nans_train, exog=exog_dict_nans_train, suppress_warnings=True
    )

    series_to_predict = ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005']
    last_window = pd.DataFrame(
        {k: v for k, v in forecaster.last_window_.items() if k in series_to_predict}
    )
    last_window['id_1005'] = last_window['id_1004']
    exog_dict_nans_test_2 = exog_dict_nans_test.copy()
    exog_dict_nans_test_2['id_1005'] = exog_dict_nans_test_2['id_1004']
    results = forecaster._create_predict_inputs(
        steps=5, levels=levels, last_window=last_window, exog=exog_dict_nans_test_2
    )
    
    expected = (
        pd.DataFrame(
            {'id_1000': np.array([-0.3525861, -0.457091, -0.49618465, -1.07810218, -1.77580056]),
             'id_1001': np.array([0.21800529, 0.36936645, 0.67896814, 1.16332175, 1.1523137]),
             'id_1003': np.array([-0.62651976, -0.74685959, -1.03823091, -0.37837738, 3.39980134]),
             'id_1004': np.array([0.61495753, 0.8322971, 0.6719899, -0.24309812, -0.75073762]),
             'id_1005': np.array([2.38367241, 2.56743042, 2.4318926, 1.65819658, 1.22899343])},
            index = pd.date_range(start='2016-07-27', periods=5, freq='D')
        ),
        {1: np.array([[0.00821644, 1.42962482, np.nan, np.nan],
                      [0.00821644, 1.42962482, 1.11141113, -0.87943526],
                      [0.00821644, np.nan, 1.11141113, -0.87943526],
                      [0.00821644, 1.42962482, 1.11141113, -0.87943526],
                      [0.00821644, 1.42962482, 1.11141113, -0.87943526]]),
         2: np.array([[1.11220226, 0.89634375, np.nan, np.nan],
                      [1.11220226, 0.89634375, 1.1327558 , 0.0058948 ],
                      [1.11220226, np.nan, 1.1327558 , 0.0058948 ],
                      [1.11220226, 0.89634375, 1.1327558 , 0.0058948 ],
                      [1.11220226, 0.89634375, 1.1327558 , 0.0058948 ]]),
         3: np.array([[1.38486425, -0.30192795, np.nan, np.nan],
                      [1.38486425, -0.30192795, 1.1775869 , -0.3532584 ],
                      [1.38486425, np.nan, 1.1775869 , -0.3532584 ],
                      [1.38486425, -0.30192795, 1.1775869 , -0.3532584 ],
                      [1.38486425, -0.30192795, 1.1775869 , -0.3532584 ]]),
         4: np.array([[0.62088235, -1.26286725, np.nan, np.nan],
                      [0.62088235, -1.26286725, 1.0428337 , 0.84287284],
                      [0.62088235, np.nan, 1.0428337 , 0.84287284],
                      [0.62088235, -1.26286725, 1.0428337 , 0.84287284],
                      [0.62088235, -1.26286725, 1.0428337 , 0.84287284]]),
         5: np.array([[-0.60444947, -1.26286725, np.nan, np.nan],
                      [-0.60444947, -1.26286725, 1.00599776, -0.62314633],
                      [-0.60444947, np.nan, 1.00599776, -0.62314633],
                      [-0.60444947, -1.26286725, 1.00599776, -0.62314633],
                      [-0.60444947, -1.26286725, 1.00599776, -0.62314633]])
        },
        ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005'],
        pd.date_range(start='2016-08-01', periods=5, freq='D')
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    for k in results[1].keys():
        np.testing.assert_array_almost_equal(results[1][k], expected[1][k])
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


@pytest.mark.parametrize("levels", 
                         [None, ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005']], 
                         ids = lambda levels: f'levels: {levels}')
def test_create_predict_inputs_output_when_series_and_exog_dict_unknown_level_encoding_None(levels):
    """
    Test output ForecasterRecursiveMultiSeries _create_predict_inputs method when 
    series and exog are dictionaries and unknown level with encoding=None.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        regressor          = LGBMRegressor(
            n_estimators=2, random_state=123, verbose=-1, max_depth=2
        ),
        lags               = 5,
        encoding           = None,
        dropna_from_series = False,
        transformer_series = StandardScaler(),
        transformer_exog   = StandardScaler(),
    )
    forecaster.fit(
        series=series_dict_nans_train, exog=exog_dict_nans_train, suppress_warnings=True
    )

    series_to_predict = ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005']
    last_window = pd.DataFrame(
        {k: v for k, v in forecaster.last_window_.items() if k in series_to_predict}
    )
    last_window['id_1005'] = last_window['id_1004']
    results = forecaster._create_predict_inputs(
        steps=5, levels=levels, last_window=last_window, exog=exog_dict_nans_test
    )
    
    expected = (
        pd.DataFrame(
            {'id_1000': np.array([-0.77768307, -0.79277912, -0.79842631, -0.88248605, -0.98327068]),
             'id_1001': np.array([-0.39572285, -0.34793695, -0.25019327, -0.09727905, -0.10075438]),
             'id_1003': np.array([-0.56650418, -0.59023234, -0.64768383, -0.51757641,  0.22739054]),
             'id_1004': np.array([2.38367241, 2.56743042, 2.4318926, 1.65819658, 1.22899343]),
             'id_1005': np.array([2.38367241, 2.56743042, 2.4318926, 1.65819658, 1.22899343])},
            index = pd.date_range(start='2016-07-27', periods=5, freq='D')
        ),
        {1: np.array([[0.00821644, 1.42962482, np.nan, np.nan],
                      [0.00821644, 1.42962482, 1.11141113, -0.87943526],
                      [0.00821644, np.nan, 1.11141113, -0.87943526],
                      [0.00821644, 1.42962482, 1.11141113, -0.87943526],
                      [np.nan, np.nan, np.nan, np.nan]]),
         2: np.array([[1.11220226, 0.89634375, np.nan, np.nan],
                      [1.11220226, 0.89634375, 1.1327558 , 0.0058948 ],
                      [1.11220226, np.nan, 1.1327558 , 0.0058948 ],
                      [1.11220226, 0.89634375, 1.1327558 , 0.0058948 ],
                      [np.nan, np.nan, np.nan, np.nan]]),
         3: np.array([[1.38486425, -0.30192795, np.nan, np.nan],
                      [1.38486425, -0.30192795, 1.1775869 , -0.3532584 ],
                      [1.38486425, np.nan, 1.1775869 , -0.3532584 ],
                      [1.38486425, -0.30192795, 1.1775869 , -0.3532584 ],
                      [np.nan, np.nan, np.nan, np.nan]]),
         4: np.array([[0.62088235, -1.26286725, np.nan, np.nan],
                      [0.62088235, -1.26286725, 1.0428337 , 0.84287284],
                      [0.62088235, np.nan, 1.0428337 , 0.84287284],
                      [0.62088235, -1.26286725, 1.0428337 , 0.84287284],
                      [np.nan, np.nan, np.nan, np.nan]]),
         5: np.array([[-0.60444947, -1.26286725, np.nan, np.nan],
                      [-0.60444947, -1.26286725, 1.00599776, -0.62314633],
                      [-0.60444947, np.nan, 1.00599776, -0.62314633],
                      [-0.60444947, -1.26286725, 1.00599776, -0.62314633],
                      [np.nan, np.nan, np.nan, np.nan]])
        },
        ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005'],
        pd.date_range(start='2016-08-01', periods=5, freq='D')
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    for k in results[1].keys():
        np.testing.assert_array_almost_equal(results[1][k], expected[1][k])
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])



# Unit test initialize_lags_grid
# ==============================================================================
import re
import pytest
import numpy as np
from sklearn.linear_model import Ridge
from skforecast.recursive import ForecasterRecursive
from skforecast.model_selection._utils import initialize_lags_grid


def test_TypeError_initialize_lags__rid_when_not_list_dict_or_None():
    """
    Test TypeError is raised when lags_grid is not a list, dict or None.
    """
    forecaster = ForecasterRecursive(regressor=Ridge(random_state=123), lags=2)
    lags_grid = 'not_valid_type'

    err_msg = re.escape(
        (f"`lags_grid` argument must be a list, dict or None. "
         f"Got {type(lags_grid)}.")
    )
    with pytest.raises(TypeError, match = err_msg):
        initialize_lags_grid(forecaster, lags_grid)


def test_initialize_lags_grid_when_lags_grid_is_a_list():
    """
    Test initialize_lags_grid when lags_grid is a list.
    """
    forecaster = ForecasterRecursive(regressor=Ridge(random_state=123), lags=2)
    lags_grid = [1, [2, 4], range(3, 5), np.array([3, 7])]
    lags_grid, lags_label = initialize_lags_grid(forecaster, lags_grid)

    lags_grid_expected = {
        '1': 1, 
        '[2, 4]': [2, 4], 
        'range(3, 5)': range(3, 5), 
        '[3 7]': np.array([3, 7])
    }

    assert lags_label == 'values'
    assert lags_grid.keys() == lags_grid_expected.keys()
    for v, v_expected in zip(lags_grid.values(), lags_grid_expected.values()):
        if isinstance(v, np.ndarray):
            assert np.array_equal(v, v_expected)
        elif isinstance(v, range):
            assert list(v) == list(v_expected)
        else:
            assert v == v_expected


@pytest.mark.parametrize("lags, lags_grid_expected",
                         [(3, {'[1, 2, 3]': [1, 2, 3]}), 
                          ([1, 2, 3], {'[1, 2, 3]': [1, 2, 3]}),
                          (range(1, 4), {'[1, 2, 3]': [1, 2, 3]}),
                          (np.array([1, 2, 3]), {'[1, 2, 3]': [1, 2, 3]})],
                         ids=lambda lags: f'lags, lags_grid_expected: {lags}')
def test_initialize_lags_grid_when_lags_grid_is_None(lags, lags_grid_expected):
    """
    Test initialize_lags_grid when lags_grid is None.
    """
    forecaster = ForecasterRecursive(regressor=Ridge(random_state=123), lags=lags)
    lags_grid = None
    lags_grid, lags_label = initialize_lags_grid(forecaster, lags_grid)

    assert lags_label == 'values'
    assert lags_grid.keys() == lags_grid_expected.keys()
    for v, v_expected in zip(lags_grid.values(), lags_grid_expected.values()):
        assert v == v_expected


def test_initialize_lags_grid_when_lags_grid_is_a_dict():
    """
    Test initialize_lags_grid when lags_grid is a dict.
    """
    forecaster = ForecasterRecursive(regressor=Ridge(random_state=123), lags=2)
    lags_grid = {'1': 1, '[2, 4]': [2, 4], 'range(3, 5)': range(3, 5), '[3 7]': np.array([3, 7])}
    lags_grid, lags_label = initialize_lags_grid(forecaster, lags_grid)

    lags_grid_expected = {
        '1': 1, 
        '[2, 4]': [2, 4], 
        'range(3, 5)': range(3, 5), 
        '[3 7]': np.array([3, 7])
    }
    
    assert lags_label == 'keys'
    assert lags_grid.keys() == lags_grid_expected.keys()
    for v, v_expected in zip(lags_grid.values(), lags_grid_expected.values()):
        if isinstance(v, np.ndarray):
            assert np.array_equal(v, v_expected)
        elif isinstance(v, range):
            assert list(v) == list(v_expected)
        else:
            assert v == v_expected


# Unit test grid_search_sarimax
# ==============================================================================
import numpy as np
import pandas as pd
from skforecast.sarimax import Sarimax
from skforecast.recursive import ForecasterSarimax
from skforecast.model_selection._split import TimeSeriesFold
from skforecast.model_selection._search import grid_search_sarimax

# Fixtures
from ....recursive.tests.tests_forecaster_sarimax.fixtures_forecaster_sarimax import y_datetime

from tqdm import tqdm
from functools import partialmethod
tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)  # hide progress bar


def test_output_grid_search_sarimax_sarimax_with_mocked():
    """
    Test output of grid_search_sarimax in ForecasterSarimax with mocked
    (mocked done in Skforecast v0.7.0).
    """
    forecaster = ForecasterSarimax(
                     regressor = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y_datetime) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    param_grid = [{'order': [(2, 2, 0), (3, 2, 0)],
                   'trend': [None, 'c']}]

    results = grid_search_sarimax(
                  forecaster  = forecaster,
                  y           = y_datetime,
                  cv          = cv,
                  param_grid  = param_grid,
                  metric      = 'mean_absolute_error',
                  return_best = False,
                  verbose     = False
              )
    
    expected_results = pd.DataFrame(
        data  = {'params'             : [{'order': (3, 2, 0), 'trend': None},
                                         {'order': (3, 2, 0), 'trend': 'c'}, 
                                         {'order': (2, 2, 0), 'trend': 'c'}, 
                                         {'order': (2, 2, 0), 'trend': None}],
                 'mean_absolute_error': np.array([0.15357204, 0.1548934 , 0.19852912, 0.19853423]),
                 'order'              : [(3, 2, 0), (3, 2, 0), (2, 2, 0), (2, 2, 0)],
                 'trend'              : [None, 'c', 'c', None]},
        index = pd.Index(np.array([0, 1, 2, 3]), dtype="int64")
    )

    pd.testing.assert_frame_equal(results, expected_results, atol=0.0001)



# Unit test predict_bootstrapping ForecasterDirect
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler
from lightgbm import LGBMRegressor

from skforecast.preprocessing import RollingFeatures
from skforecast.preprocessing import TimeSeriesDifferentiator
from skforecast.direct import ForecasterDirect

# Fixtures
from .fixtures_forecaster_direct import y
from .fixtures_forecaster_direct import exog
from .fixtures_forecaster_direct import exog_predict
from .fixtures_forecaster_direct import data  # to test results when using differentiation


def test_predict_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=5)

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict_bootstrapping(steps=5)


@pytest.mark.parametrize("use_binned_residuals", [True, False], 
                         ids=lambda binned: f'use_binned_residuals: {binned}')
def test_predict_bootstrapping_ValueError_when_out_sample_residuals_is_None(use_binned_residuals):
    """
    Test ValueError is raised when use_in_sample_residuals=False and
    out sample residuals is None.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=2)
    forecaster.fit(y=pd.Series(np.arange(10)), store_in_sample_residuals=True)

    if use_binned_residuals:
        literal = "out_sample_residuals_by_bin_"
    else:
        literal = "out_sample_residuals_"

    err_msg = re.escape(
        f"`forecaster.{literal}` is either None or empty. Use "
        f"`use_in_sample_residuals = True` or the `set_out_sample_residuals()` "
        f"method before predicting."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(
            steps=1, use_in_sample_residuals=False, use_binned_residuals=use_binned_residuals
        )


@pytest.mark.parametrize("steps", [2, [1, 2], None], 
                         ids=lambda steps: f'steps: {steps}')
def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer(steps):
    """
    Test output of predict_bootstrapping when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterDirect(
                     regressor        = LinearRegression(),
                     steps            = 2,
                     lags             = 3,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                 )
    forecaster.fit(y=y, exog=exog, store_in_sample_residuals=True)
    results = forecaster.predict_bootstrapping(
        steps=steps, exog=exog_predict, n_boot=4, 
        use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data = np.array([[0.83491717, 0.72256471, 0.6291488 , 0.75497624],
                                    [0.0844494 , 0.51730407, 0.30032917, 0.41397707]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_False_exog_and_transformer():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterDirect(
                     regressor        = LinearRegression(),
                     steps            = 2,
                     lags             = 3,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                 )
    forecaster.fit(y=y, exog=exog, store_in_sample_residuals=True)
    forecaster.out_sample_residuals_ = forecaster.in_sample_residuals_
    results = forecaster.predict_bootstrapping(
        steps=2, exog=exog_predict, n_boot=4, 
        use_in_sample_residuals=False, use_binned_residuals=False
    )
    
    expected = pd.DataFrame(
                   data = np.array([[0.83491717, 0.72256471, 0.6291488 , 0.75497624],
                                    [0.0844494 , 0.51730407, 0.30032917, 0.41397707]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_fixed():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals that are fixed.
    """
    forecaster = ForecasterDirect(
                     regressor = LinearRegression(),
                     steps     = 2,
                     lags      = 3
                 )
    forecaster.fit(y=y, exog=exog, store_in_sample_residuals=True)
    forecaster.in_sample_residuals_ = np.array(
        [1, 1, 1, 1, 1, 1, 1, 5, 5, 5, 5, 5, 5, 5]
    )
    results = forecaster.predict_bootstrapping(
        steps=2, exog=exog_predict, n_boot=4, 
        use_in_sample_residuals=True, use_binned_residuals=False
    )
    
    expected = pd.DataFrame(
                   data = np.array([[1.67523588, 5.67523588, 5.67523588, 1.67523588],
                                    [5.38024988, 1.38024988, 1.38024988, 1.38024988]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_and_differentiation_is_1_steps_1():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression and
    differentiation=1 and steps=1.
    """
    # Data differentiated
    differentiator = TimeSeriesDifferentiator(order=1)
    data_diff = differentiator.fit_transform(data.to_numpy())
    data_diff = pd.Series(data_diff, index=data.index).dropna()

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff = exog.iloc[1:]
    end_train = '2003-03-01 23:59:00'

    forecaster_1 = ForecasterDirect(regressor=LinearRegression(), steps=1, lags=15)
    forecaster_1.fit(
        y=data_diff.loc[:end_train], exog=exog_diff.loc[:end_train], store_in_sample_residuals=True
    )
    boot_predictions_diff = forecaster_1.predict_bootstrapping(
        exog=exog_diff.loc[end_train:], n_boot=10, use_in_sample_residuals=True, use_binned_residuals=False
    )

    # Revert the differentiation
    last_value_train = data.loc[:end_train].iloc[[-1]]
    boot_predictions_1 = boot_predictions_diff.copy()
    boot_predictions_1.loc[last_value_train.index[0]] = last_value_train.values[0]
    boot_predictions_1 = boot_predictions_1.sort_index()
    boot_predictions_1 = boot_predictions_1.cumsum(axis=0).iloc[1:,]
    boot_predictions_1 = boot_predictions_1.asfreq('MS')
    
    forecaster_2 = ForecasterDirect(regressor=LinearRegression(), steps=1, lags=15, differentiation=1)
    forecaster_2.fit(
        y=data.loc[:end_train], exog=exog.loc[:end_train], store_in_sample_residuals=True
    )
    boot_predictions_2 = forecaster_2.predict_bootstrapping(
        exog=exog_diff.loc[end_train:], n_boot=10, use_in_sample_residuals=True, use_binned_residuals=False
    )
    
    pd.testing.assert_frame_equal(boot_predictions_1, boot_predictions_2)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_and_differentiation_is_1_steps_10():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression and
    differentiation=1 and steps=10.
    """
    # Data differentiated
    differentiator = TimeSeriesDifferentiator(order=1)
    data_diff = differentiator.fit_transform(data.to_numpy())
    data_diff = pd.Series(data_diff, index=data.index).dropna()

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff = exog.iloc[1:]
    end_train = '2003-03-01 23:59:00'

    forecaster_1 = ForecasterDirect(regressor=LinearRegression(), steps=10, lags=15)
    forecaster_1.fit(
        y=data_diff.loc[:end_train], exog=exog_diff.loc[:end_train], store_in_sample_residuals=True
    )
    boot_predictions_diff = forecaster_1.predict_bootstrapping(
        exog=exog_diff.loc[end_train:], n_boot=10, use_in_sample_residuals=True, use_binned_residuals=False
    )
    last_value_train = data.loc[:end_train].iloc[[-1]]
    boot_predictions_1 = boot_predictions_diff.copy()
    boot_predictions_1.loc[last_value_train.index[0]] = last_value_train.values[0]
    boot_predictions_1 = boot_predictions_1.sort_index()
    boot_predictions_1 = boot_predictions_1.cumsum(axis=0).iloc[1:,]
    boot_predictions_1 = boot_predictions_1.asfreq('MS')
    
    forecaster_2 = ForecasterDirect(regressor=LinearRegression(), steps=10, lags=15, differentiation=1)
    forecaster_2.fit(
        y=data.loc[:end_train], exog=exog.loc[:end_train], store_in_sample_residuals=True
    )
    boot_predictions_2 = forecaster_2.predict_bootstrapping(
        exog=exog_diff.loc[end_train:], n_boot=10, use_in_sample_residuals=True, use_binned_residuals=False
    )
    
    pd.testing.assert_frame_equal(boot_predictions_1, boot_predictions_2)


def test_predict_bootstrapping_output_when_window_features_steps_1():
    """
    Test output of predict_bootstrapping when regressor is LGBMRegressor and
    4 steps ahead are predicted with exog and window features using in-sample residuals
    with steps=1.
    """
    y_datetime = y.copy()
    y_datetime.index = pd.date_range(start='2001-01-01', periods=len(y), freq='D')
    exog_datetime = exog.copy()
    exog_datetime.index = pd.date_range(start='2001-01-01', periods=len(exog), freq='D')
    exog_predict_datetime = exog_predict.copy()
    exog_predict_datetime.index = pd.date_range(start='2001-02-20', periods=len(exog_predict), freq='D')
    
    rolling = RollingFeatures(stats=['mean', 'sum'], window_sizes=[3, 5])
    forecaster = ForecasterDirect(
        LGBMRegressor(verbose=-1, random_state=123), steps=1, lags=3, window_features=rolling
    )
    forecaster.fit(y=y_datetime, exog=exog_datetime, store_in_sample_residuals=True)
    results = forecaster.predict_bootstrapping(
        n_boot=10, exog=exog_predict_datetime, 
        use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data    = np.array(
                                 [[0.61866004, 0.42315364, 0.54459359, 0.6327955 , 0.29239436,
                                   0.46729748, 0.3783451 , 0.18128852, 0.71765599, 0.58209158]]
                             ),
                   columns = [f"pred_boot_{i}" for i in range(10)],
                   index   = pd.date_range(start='2001-02-20', periods=1, freq='D')
               )

    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_window_features_steps_10():
    """
    Test output of predict_bootstrapping when regressor is LGBMRegressor and
    4 steps ahead are predicted with exog and window features using in-sample residuals
    with steps=10.
    """
    y_datetime = y.copy()
    y_datetime.index = pd.date_range(start='2001-01-01', periods=len(y), freq='D')
    exog_datetime = exog.copy()
    exog_datetime.index = pd.date_range(start='2001-01-01', periods=len(exog), freq='D')
    exog_predict_datetime = exog_predict.copy()
    exog_predict_datetime.index = pd.date_range(start='2001-02-20', periods=len(exog_predict), freq='D')
    
    rolling = RollingFeatures(stats=['mean', 'sum'], window_sizes=[3, 5])
    forecaster = ForecasterDirect(
        LGBMRegressor(verbose=-1, random_state=123), steps=10, lags=3, window_features=rolling
    )
    forecaster.fit(y=y_datetime, exog=exog_datetime, store_in_sample_residuals=True)
    results = forecaster.predict_bootstrapping(
        n_boot=10, exog=exog_predict_datetime, 
        use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data    = np.array(
                                 [[0.34317802, 0.64776959, 0.44252774, 0.72244338, 0.17452841,
                                    0.42501556, 0.38875996, 0.43489265, 0.66887236, 0.50222644],
                                [0.34888897, 0.1813542 , 0.24348837, 0.71499063, 0.3770138 ,
                                    0.44426025, 0.42297623, 0.6403034 , 0.31816366, 0.25635781],
                                [0.37107296, 0.39877449, 0.43370117, 0.61102351, 0.84943179,
                                    0.83690427, 0.24017543, 0.72904971, 0.43536741, 0.35150872],
                                [0.58509345, 0.83877703, 0.7137504 , 0.57562876, 0.34400862,
                                    0.52405127, 0.30293129, 0.96777975, 0.72154961, 0.17572187],
                                [0.24338025, 0.5347562 , 0.61422834, 0.61575098, 0.66106327,
                                    0.86630916, 0.31941155, 0.50577174, 0.39804426, 0.41112617],
                                [0.0988748 , 0.31547649, 0.83027159, 0.42189523, 0.40316123,
                                    0.71344515, 0.227533  , 0.60577608, 0.41454097, 0.63051607],
                                [0.7418803 , 0.60966097, 0.17545176, 0.41697845, 0.4308126 ,
                                    0.06283256, 0.92888472, 0.41697845, 0.48408656, 0.60389823],
                                [0.54283776, 0.32254116, 0.44564736, 0.48303426, 0.30078917,
                                    0.42410373, 0.07449023, 0.32756542, 0.54283776, 0.54283776],
                                [0.33621667, 0.07766116, 0.51371272, 0.51509444, 0.61903383,
                                    0.64692848, 0.37116152, 0.3145088 , 0.70908819, 0.42496032],
                                [0.91812915, 0.47396699, 0.63571333, 0.34410137, 0.65201567,
                                    0.18341508, 0.46122504, 0.43951603, 0.33700121, 0.51837492]]
                             ),
                   columns = [f"pred_boot_{i}" for i in range(10)],
                   index   = pd.date_range(start='2001-02-20', periods=10, freq='D')
               )

    pd.testing.assert_frame_equal(expected, results)



# Unit test _create_train_X_y_single_series ForecasterRecursiveMultiSeries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.base import clone
from sklearn.linear_model import LinearRegression
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import MinMaxScaler
from skforecast.preprocessing import RollingFeatures
from ....recursive import ForecasterRecursiveMultiSeries


def test_create_train_X_y_single_series_ValueError_when_len_y_less_than_window_size():
    """
    Test ValueError is raised when len(y) <= window_size.
    """
    y = pd.Series(np.arange(5), name='l1')
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)

    err_msg = re.escape(
        "Length of 'l1' must be greater than the maximum window size "
        "needed by the forecaster.\n"
        "    Length 'l1': 5.\n"
        "    Max window size: 5.\n"
        "    Lags window size: 5.\n"
        "    Window features window size: None."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y_single_series(y=y, ignore_exog=True)

    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=6)
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=2, window_features=rolling)
    err_msg = re.escape(
        "Length of 'l1' must be greater than the maximum window size "
        "needed by the forecaster.\n"
        "    Length 'l1': 5.\n"
        "    Max window size: 6.\n"
        "    Lags window size: 2.\n"
        "    Window features window size: 6."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y_single_series(y=y, ignore_exog=True)


@pytest.mark.parametrize("ignore_exog", 
                         [True, False], 
                         ids = lambda ie: f'ignore_exog: {ie}')
def test_create_train_X_y_single_series_output_when_series_and_exog_is_None(ignore_exog):
    """
    Test the output of _create_train_X_y_single_series when exog is None. 
    Check cases `ignore_exog` True and False.
    """
    y = pd.Series(np.arange(7, dtype=float), name='l1')
    
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.transformer_series_ = {'l1': StandardScaler()}
    forecaster.differentiator_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = ignore_exog,
                  exog        = None
              )
    
    expected = (
        pd.DataFrame(
            data = np.array([[-0.5, -1. , -1.5, 'l1'],
                             [ 0. , -0.5, -1. , 'l1'],
                             [ 0.5,  0. , -0.5, 'l1'],
                             [ 1. ,  0.5,  0. , 'l1']]),
            index   = pd.RangeIndex(start=3, stop=7, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float}),
        None,
        pd.Series(
            data  = np.nan,
            name  = '_dummy_exog_col_to_keep_shape',
            index = pd.RangeIndex(start=3, stop=7, step=1)
        ),
        pd.Series(
            data  = np.array([0., 0.5, 1., 1.5]),
            index = pd.RangeIndex(start=3, stop=7, step=1),
            name  = 'y',
            dtype = float
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], type(None))
    if ignore_exog:
        assert isinstance(results[2], type(None))
    else:
        pd.testing.assert_series_equal(results[2], expected[2])
    pd.testing.assert_series_equal(results[3], expected[3])


def test_create_train_X_y_single_series_output_when_series_and_exog():
    """
    Test the output of _create_train_X_y_single_series when exog is a 
    pandas DataFrame with one column.
    """
    y = pd.Series(np.arange(10, dtype=float), name='l1')
    exog = pd.DataFrame(np.arange(100, 110, dtype=float), columns=['exog'])
    
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5,
                                                transformer_series=None)
    forecaster.transformer_series_ = {'l1': None}
    forecaster.differentiator_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 'l1'],
                             [5., 4., 3., 2., 1., 'l1'],
                             [6., 5., 4., 3., 2., 'l1'],
                             [7., 6., 5., 4., 3., 'l1'],
                             [8., 7., 6., 5., 4., 'l1']]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float,
                  'lag_4': float, 'lag_5': float}),
        None,
        pd.DataFrame(
            data  = np.array([105., 106., 107., 108., 109.], dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['exog']
        ),
        pd.Series(
            data  = np.array([5., 6., 7., 8., 9.], dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], type(None))
    pd.testing.assert_frame_equal(results[2], expected[2])
    pd.testing.assert_series_equal(results[3], expected[3])


def test_create_train_X_y_single_series_output_when_series_10_and_exog_is_dataframe_of_category():
    """
    Test the output of _create_train_X_y_single_series when exog is a 
    pandas DataFrame with one column float and one column of cateogry.
    """
    y = pd.Series(np.arange(10, dtype=float), name='l1')
    exog = pd.DataFrame({'exog_1': np.arange(10, 20, dtype=float),
                         'exog_2': pd.Categorical(range(100, 110))})

    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5,
                                                transformer_series=None)
    forecaster.transformer_series_ = {'l1': None}
    forecaster.differentiator_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 'l1'],
                             [5., 4., 3., 2., 1., 'l1'],
                             [6., 5., 4., 3., 2., 'l1'],
                             [7., 6., 5., 4., 3., 'l1'],
                             [8., 7., 6., 5., 4., 'l1']]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float,
                  'lag_4': float, 'lag_5': float}),
        None,
        pd.DataFrame(
            data  = np.array([15., 16., 17., 18., 19.], dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['exog_1']
        ).assign(exog_2 = pd.Categorical([105, 106, 107, 108, 109], categories=range(100, 110))),
        pd.Series(
            data  = np.array([5., 6., 7., 8., 9.], dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], type(None))
    pd.testing.assert_frame_equal(results[2], expected[2])
    pd.testing.assert_series_equal(results[3], expected[3])


def test_create_train_X_y_single_series_output_when_series_and_exog_is_DataFrame_datetime_index():
    """
    Test the output of _create_train_X_y_single_series when series and 
    exog is a pandas dataframe with two columns and datetime index.
    """
    y = pd.Series(np.arange(7, dtype=float), name='l1')
    y.index = pd.date_range("1990-01-01", periods=7, freq='D')
    exog = pd.DataFrame({
               'exog_1': np.arange(100, 107, dtype=float),
               'exog_2': np.arange(1000, 1007, dtype=float)},
               index = pd.date_range("1990-01-01", periods=7, freq='D')
           )

    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3,
                                                transformer_series=None)
    forecaster.transformer_series_ = {'l1': None}
    forecaster.differentiator_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )
    
    expected = (
        pd.DataFrame(
            data = np.array([[2.0, 1.0, 0.0, 'l1'],
                             [3.0, 2.0, 1.0, 'l1'],
                             [4.0, 3.0, 2.0, 'l1'],
                             [5.0, 4.0, 3.0, 'l1']]),
            index   = pd.date_range("1990-01-04", periods=4, freq='D'),
            columns = ['lag_1', 'lag_2', 'lag_3', '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float}),
        None,
        pd.DataFrame(
            data = np.array([[103., 1003.],
                             [104., 1004.],
                             [105., 1005.],
                             [106., 1006.]]),
            index   = pd.date_range("1990-01-04", periods=4, freq='D'),
            columns = ['exog_1', 'exog_2']
        ),
        pd.Series(
            data  = np.array([3., 4., 5., 6.], dtype=float),
            index = pd.date_range("1990-01-04", periods=4, freq='D'),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], type(None))
    pd.testing.assert_frame_equal(results[2], expected[2])
    pd.testing.assert_series_equal(results[3], expected[3])


def test_create_train_X_y_single_series_output_when_series_and_exog_is_DataFrame_with_NaNs():
    """
    Test the output of _create_train_X_y_single_series when series and 
    exog is a pandas dataframe with two columns and NaNs in between.
    """
    y = pd.Series(np.arange(10, dtype=float), name='l1')
    y.iloc[6:8] = np.nan
    exog = pd.DataFrame({
               'exog_1': np.arange(100, 110, dtype=float),
               'exog_2': np.arange(1000, 1010, dtype=float)}
           )
    exog.iloc[2:7, 0] = np.nan

    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5,
                                                transformer_series=None)
    forecaster.transformer_series_ = {'l1': None}
    forecaster.differentiator_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 'l1'],
                             [5., 4., 3., 2., 1., 'l1'],
                             [np.nan, 5., 4., 3., 2., 'l1'],
                             [np.nan, np.nan, 5., 4., 3., 'l1'],
                             [8., np.nan, np.nan, 5., 4., 'l1']]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float,
                  'lag_4': float, 'lag_5': float}),
        None,
        pd.DataFrame(
            data = np.array([[np.nan, 1005.],
                             [np.nan, 1006.],
                             [107., 1007.],
                             [108., 1008.],
                             [109., 1009.]]),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['exog_1', 'exog_2']
        ),
        pd.Series(
            data  = np.array([5., np.nan, np.nan, 8., 9.], dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], type(None))
    pd.testing.assert_frame_equal(results[2], expected[2])
    pd.testing.assert_series_equal(results[3], expected[3])


def test_create_train_X_y_single_series_output_when_transformer_and_fitted():
    """
    Test the output of _create_train_X_y_single_series when Forecaster as
    already been fitted, transformer is MinMaxScaler() and has a different 
    series as input.
    """
    y = pd.Series(np.arange(9, dtype=float), name='l1')
    transformer = MinMaxScaler()
    transformer.fit(y.values.reshape(-1, 1))

    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3,
                                                transformer_series=MinMaxScaler())
    forecaster.transformer_series_ = {'l1': transformer}
    forecaster.differentiator_ = {'l1': None}
    forecaster.is_fitted = True

    new_y = pd.Series(np.arange(10, 19, dtype=float), name='l1')
    results = forecaster._create_train_X_y_single_series(y=new_y, ignore_exog=True)
    
    expected = (
        pd.DataFrame(
            data = np.array([[1.5  , 1.375, 1.25 , 'l1'],
                             [1.625, 1.5  , 1.375, 'l1'],
                             [1.75 , 1.625, 1.5  , 'l1'],
                             [1.875, 1.75 , 1.625, 'l1'],
                             [2.   , 1.875, 1.75 , 'l1'],
                             [2.125, 2.   , 1.875, 'l1']]),
            index   = pd.RangeIndex(start=3, stop=9, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float}),
        None,
        None,
        pd.Series(
            data  = np.array([1.625, 1.75, 1.875, 2., 2.125, 2.25]),
            index = pd.RangeIndex(start=3, stop=9, step=1),
            name  = 'y',
            dtype = float
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], type(None))
    assert isinstance(results[2], type(None))
    pd.testing.assert_series_equal(results[3], expected[3])


@pytest.mark.parametrize("is_fitted", 
                         [True, False], 
                         ids = lambda is_fitted: f'is_fitted: {is_fitted}')
def test_create_train_X_y_single_series_output_when_series_and_exog_and_differentitation_1(is_fitted):
    """
    Test the output of _create_train_X_y_single_series when exog is a 
    pandas DataFrame with one column and differentiation=1.
    """
    y = pd.Series(np.arange(10, dtype=float), name='l1')
    exog = pd.DataFrame(np.arange(100, 110, dtype=float), columns=['exog'])
    
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5,
                                                transformer_series = None,
                                                differentiation    = 1)
    forecaster.transformer_series_ = {'l1': None}
    forecaster.differentiator_ = {'l1': clone(forecaster.differentiator)}
    forecaster.is_fitted = is_fitted

    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )

    expected = (
        pd.DataFrame(
            data = np.array([[1., 1., 1., 1., 1., 'l1'],
                             [1., 1., 1., 1., 1., 'l1'],
                             [1., 1., 1., 1., 1., 'l1'],
                             [1., 1., 1., 1., 1., 'l1']]),
            index   = pd.RangeIndex(start=6, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float,
                  'lag_4': float, 'lag_5': float}),
        None,
        pd.DataFrame(
            data  = np.array([106., 107., 108., 109.], dtype=float),
            index = pd.RangeIndex(start=6, stop=10, step=1),
            columns = ['exog']
        ),
        pd.Series(
            data  = np.array([1., 1., 1., 1.], dtype=float),
            index = pd.RangeIndex(start=6, stop=10, step=1),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], type(None))
    pd.testing.assert_frame_equal(results[2], expected[2])
    pd.testing.assert_series_equal(results[3], expected[3])


@pytest.mark.parametrize("is_fitted", 
                         [True, False], 
                         ids = lambda is_fitted: f'is_fitted: {is_fitted}')
def test_create_train_X_y_single_series_output_when_series_and_exog_and_differentitation_2(is_fitted):
    """
    Test the output of _create_train_X_y_single_series when exog is a 
    pandas DataFrame with one column and differentiation=2.
    """
    y = pd.Series(np.arange(10, dtype=float), name='l1')
    exog = pd.DataFrame(np.arange(100, 110, dtype=float), columns=['exog'])
    
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5,
                                                transformer_series = None,
                                                differentiation    = 2)
    forecaster.transformer_series_ = {'l1': None}
    forecaster.differentiator_ = {'l1': clone(forecaster.differentiator)}
    forecaster.is_fitted = is_fitted

    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )

    expected = (
        pd.DataFrame(
            data = np.array([[0., 0., 0., 0., 0., 'l1'],
                             [0., 0., 0., 0., 0., 'l1'],
                             [0., 0., 0., 0., 0., 'l1']]),
            index   = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float,
                  'lag_4': float, 'lag_5': float}),
        None,
        pd.DataFrame(
            data  = np.array([107., 108., 109.], dtype=float),
            index = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['exog']
        ),
        pd.Series(
            data  = np.array([0., 0., 0.], dtype=float),
            index = pd.RangeIndex(start=7, stop=10, step=1),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], type(None))
    pd.testing.assert_frame_equal(results[2], expected[2])
    pd.testing.assert_series_equal(results[3], expected[3])


def test_create_train_X_y_single_series_output_when_window_features_and_exog():
    """
    Test the output of _create_train_X_y_single_series when using window_features 
    and exog with datetime index.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='l1', dtype=float
    )
    exog_datetime = pd.DataFrame(
        np.arange(100, 115, dtype=float), 
        index   = pd.date_range('2000-01-01', periods=15, freq='D'),
        columns = ['exog']
    )
    rolling = RollingFeatures(
        stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
    )

    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=5, window_features=rolling
    )
    forecaster.transformer_series_ = {'l1': None}
    forecaster.differentiator_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y_datetime,
                  ignore_exog = False,
                  exog        = exog_datetime
              )
    
    expected = (
        pd.DataFrame(
            data = np.array([[5., 4., 3., 2., 1., 3., 3., 15., 'l1'],
                             [6., 5., 4., 3., 2., 4., 4., 21., 'l1'],
                             [7., 6., 5., 4., 3., 5., 5., 27., 'l1'],
                             [8., 7., 6., 5., 4., 6., 6., 33., 'l1'],
                             [9., 8., 7., 6., 5., 7., 7., 39., 'l1'],
                             [10., 9., 8., 7., 6., 8., 8., 45., 'l1'],
                             [11., 10., 9., 8., 7., 9., 9., 51., 'l1'],
                             [12., 11., 10., 9., 8., 10., 10., 57., 'l1'],
                             [13., 12., 11., 10., 9., 11., 11., 63., 'l1']]),
            index   = pd.date_range('2000-01-07', periods=9, freq='D'),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'roll_mean_5', 'roll_median_5', 'roll_sum_6', 
                       '_level_skforecast']
        ).astype(
            {'lag_1': float, 'lag_2': float, 'lag_3': float,
             'lag_4': float, 'lag_5': float, 'roll_mean_5': float,
             'roll_median_5': float, 'roll_sum_6': float}
        ),
        ['roll_mean_5', 'roll_median_5', 'roll_sum_6'],
        pd.DataFrame(
            data  = np.arange(106, 115, dtype=float),
            index = pd.date_range('2000-01-07', periods=9, freq='D'),
            columns = ['exog']
        ),
        pd.Series(
            data  = np.array([6, 7, 8, 9, 10, 11, 12, 13, 14]),
            index = pd.date_range('2000-01-07', periods=9, freq='D'),
            name  = 'y',
            dtype = float
        ),
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]
    pd.testing.assert_frame_equal(results[2], expected[2])
    pd.testing.assert_series_equal(results[3], expected[3])


def test_create_train_X_y_single_series_output_when_two_window_features_and_exog():
    """
    Test the output of _create_train_X_y_single_series when using 2 window_features 
    and exog with datetime index.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='l1', dtype=float
    )
    exog_datetime = pd.DataFrame(
        np.arange(100, 115, dtype=float), 
        index   = pd.date_range('2000-01-01', periods=15, freq='D'),
        columns = ['exog']
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])

    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=5, window_features=[rolling, rolling_2]
    )
    forecaster.transformer_series_ = {'l1': None}
    forecaster.differentiator_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y_datetime,
                  ignore_exog = False,
                  exog        = exog_datetime
              )
    
    expected = (
        pd.DataFrame(
            data = np.array([[5., 4., 3., 2., 1., 3., 3., 15., 'l1'],
                             [6., 5., 4., 3., 2., 4., 4., 21., 'l1'],
                             [7., 6., 5., 4., 3., 5., 5., 27., 'l1'],
                             [8., 7., 6., 5., 4., 6., 6., 33., 'l1'],
                             [9., 8., 7., 6., 5., 7., 7., 39., 'l1'],
                             [10., 9., 8., 7., 6., 8., 8., 45., 'l1'],
                             [11., 10., 9., 8., 7., 9., 9., 51., 'l1'],
                             [12., 11., 10., 9., 8., 10., 10., 57., 'l1'],
                             [13., 12., 11., 10., 9., 11., 11., 63., 'l1']]),
            index   = pd.date_range('2000-01-07', periods=9, freq='D'),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'roll_mean_5', 'roll_median_5', 'roll_sum_6', 
                       '_level_skforecast']
        ).astype(
            {'lag_1': float, 'lag_2': float, 'lag_3': float,
             'lag_4': float, 'lag_5': float, 'roll_mean_5': float,
             'roll_median_5': float, 'roll_sum_6': float}
        ),
        ['roll_mean_5', 'roll_median_5', 'roll_sum_6'],
        pd.DataFrame(
            data  = np.arange(106, 115, dtype=float),
            index = pd.date_range('2000-01-07', periods=9, freq='D'),
            columns = ['exog']
        ),
        pd.Series(
            data  = np.array([6, 7, 8, 9, 10, 11, 12, 13, 14]),
            index = pd.date_range('2000-01-07', periods=9, freq='D'),
            name  = 'y',
            dtype = float
        ),
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]
    pd.testing.assert_frame_equal(results[2], expected[2])
    pd.testing.assert_series_equal(results[3], expected[3])


def test__create_train_X_y_single_series_output_when_window_features_lags_None_and_exog():
    """
    Test the output of _create_train_X_y_single_series when using window_features 
    and exog with datetime index and lags=None.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='l1', dtype=float
    )
    exog_datetime = pd.DataFrame(
        np.arange(100, 115, dtype=float), 
        index   = pd.date_range('2000-01-01', periods=15, freq='D'),
        columns = ['exog']
    )
    rolling = RollingFeatures(
        stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
    )

    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=None, window_features=rolling
    )
    forecaster.transformer_series_ = {'l1': None}
    forecaster.differentiator_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y_datetime,
                  ignore_exog = False,
                  exog        = exog_datetime
              )
    
    expected = (
        pd.DataFrame(
            data = np.array([[3., 3., 15., 'l1'],
                             [4., 4., 21., 'l1'],
                             [5., 5., 27., 'l1'],
                             [6., 6., 33., 'l1'],
                             [7., 7., 39., 'l1'],
                             [8., 8., 45., 'l1'],
                             [9., 9., 51., 'l1'],
                             [10., 10., 57., 'l1'],
                             [11., 11., 63., 'l1']]),
            index   = pd.date_range('2000-01-07', periods=9, freq='D'),
            columns = ['roll_mean_5', 'roll_median_5', 'roll_sum_6', 
                       '_level_skforecast']
        ).astype(
            {'roll_mean_5': float, 'roll_median_5': float, 'roll_sum_6': float}
        ),
        ['roll_mean_5', 'roll_median_5', 'roll_sum_6'],
        pd.DataFrame(
            data  = np.arange(106, 115, dtype=float),
            index = pd.date_range('2000-01-07', periods=9, freq='D'),
            columns = ['exog']
        ),
        pd.Series(
            data  = np.array([6, 7, 8, 9, 10, 11, 12, 13, 14]),
            index = pd.date_range('2000-01-07', periods=9, freq='D'),
            name  = 'y',
            dtype = float
        ),
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]
    pd.testing.assert_frame_equal(results[2], expected[2])
    pd.testing.assert_series_equal(results[3], expected[3])


def test_create_train_X_y_single_series_output_when_window_features_and_exog_transformers_diff():
    """
    Test the output of _create_train_X_y_single_series when using window_features, 
    exog, transformers and differentiation.
    """
    y_datetime = pd.Series(
        [25.3, 29.1, 27.5, 24.3, 2.1, 46.5, 31.3, 87.1, 133.5, 4.3],
        index=pd.date_range('2000-01-01', periods=10, freq='D'),
        name='l1', dtype=float
    )
    exog = pd.DataFrame({
               'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4, 14.6, 73.5],
               'col_2': ['a', 'a', 'a', 'a', 'a', 'b', 'b', 'b', 'b', 'b']},
               index = pd.date_range('2000-01-01', periods=10, freq='D')
           )

    transformer_series = StandardScaler()
    transformer_exog = ColumnTransformer(
                            [('scale', StandardScaler(), ['col_1']),
                             ('onehot', OneHotEncoder(), ['col_2'])],
                            remainder = 'passthrough',
                            verbose_feature_names_out = False
                        )
    rolling = RollingFeatures(
        stats=['ratio_min_max', 'median'], window_sizes=4
    )
    # Exog is not transformed as it is done in _create_train_X_y
    forecaster = ForecasterRecursiveMultiSeries(
                     LinearRegression(), 
                     lags               = [1, 5], 
                     window_features    = rolling,
                     encoding           = 'onehot',
                     transformer_series = transformer_series,
                     transformer_exog   = transformer_exog,
                     differentiation    = 2
                 )
    forecaster.transformer_series_ = {'l1': transformer_series}
    forecaster.differentiator_ = {'l1': clone(forecaster.differentiator)}

    results = forecaster._create_train_X_y_single_series(
                  y           = y_datetime,
                  ignore_exog = False,
                  exog        = exog
              )
    
    expected = (
        pd.DataFrame(
            data = np.array([[-1.56436158, -0.14173746, -0.89489489, -0.27035108, 'l1'],
                             [ 1.8635851 , -0.04199628, -0.83943662,  0.62469472, 'l1'],
                             [-0.24672817, -0.49870587, -0.83943662,  0.75068358, 'l1']]),
            index   = pd.date_range('2000-01-08', periods=3, freq='D'),
            columns = ['lag_1', 'lag_5', 'roll_ratio_min_max_4', 'roll_median_4', 
                       '_level_skforecast']
        ).astype(
            {'lag_1': float, 'lag_5': float, 
             'roll_ratio_min_max_4': float, 'roll_median_4': float}
        ),
        ['roll_ratio_min_max_4', 'roll_median_4'],
        pd.DataFrame(
            data  = np.array([[47.4, 'b'],
                              [14.6, 'b'],
                              [73.5, 'b']]),
            index = pd.date_range('2000-01-08', periods=3, freq='D'),
            columns = ['col_1', 'col_2']
        ).astype({'col_1': float}
        ),
        pd.Series(
            data  = np.array([1.8635851, -0.24672817, -4.60909217]),
            index = pd.date_range('2000-01-08', periods=3, freq='D'),
            name  = 'y',
            dtype = float
        ),
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]
    pd.testing.assert_frame_equal(results[2], expected[2])
    pd.testing.assert_series_equal(results[3], expected[3])



################################################################################
#                     skforecast.model_selection._split                        #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

from __future__ import annotations
from typing import Any
from copy import deepcopy
import warnings
import numpy as np
import pandas as pd
import itertools
import skforecast
from ..utils import date_to_index_position, get_style_repr_html
from ..exceptions import IgnoredArgumentWarning


class BaseFold():
    """
    Base class for all Fold classes in skforecast. All fold classes should specify
    all the parameters that can be set at the class level in their ``__init__``.

    Parameters
    ----------
    steps : int, default None
        Number of observations used to be predicted in each fold. This is also commonly
        referred to as the forecast horizon or test size.
    initial_train_size : int, str, pandas Timestamp, default None
        Number of observations used for initial training.

        - If an integer, the number of observations used for initial training.
        - If a date string or pandas Timestamp, it is the last date included in 
        the initial training set.
    window_size : int, default None
        Number of observations needed to generate the autoregressive predictors.
    differentiation : int, default None
        Number of observations to use for differentiation. This is used to extend the
        `last_window` as many observations as the differentiation order.
    refit : bool, int, default False
        Whether to refit the forecaster in each fold.

        - If `True`, the forecaster is refitted in each fold.
        - If `False`, the forecaster is trained only in the first fold.
        - If an integer, the forecaster is trained in the first fold and then refitted
          every `refit` folds.
    fixed_train_size : bool, default True
        Whether the training size is fixed or increases in each fold.
    gap : int, default 0
        Number of observations between the end of the training set and the start of the
        test set.
    skip_folds : int, list, default None
        Number of folds to skip.

        - If an integer, every 'skip_folds'-th is returned.
        - If a list, the indexes of the folds to skip.

        For example, if `skip_folds=3` and there are 10 folds, the returned folds are
        0, 3, 6, and 9. If `skip_folds=[1, 2, 3]`, the returned folds are 0, 4, 5, 6, 7,
        8, and 9.
    allow_incomplete_fold : bool, default True
        Whether to allow the last fold to include fewer observations than `steps`.
        If `False`, the last fold is excluded if it is incomplete.
    return_all_indexes : bool, default False
        Whether to return all indexes or only the start and end indexes of each fold.
    verbose : bool, default True
        Whether to print information about generated folds.

    Attributes
    ----------
    steps : int
        Number of observations used to be predicted in each fold. This is also commonly
        referred to as the forecast horizon or test size.
    initial_train_size : int
        Number of observations used for initial training.
    window_size : int
        Number of observations needed to generate the autoregressive predictors.
    differentiation : int
        Number of observations to use for differentiation. This is used to extend the
        `last_window` as many observations as the differentiation order.
    refit : bool, int
        Whether to refit the forecaster in each fold.
    fixed_train_size : bool
        Whether the training size is fixed or increases in each fold.
    gap : int
        Number of observations between the end of the training set and the start of the
        test set.
    skip_folds : int, list
        Number of folds to skip.
    allow_incomplete_fold : bool
        Whether to allow the last fold to include fewer observations than `steps`.
    return_all_indexes : bool
        Whether to return all indexes or only the start and end indexes of each fold.
    verbose : bool
        Whether to print information about generated folds.

    """

    def __init__(
        self,
        steps: int | None = None,
        initial_train_size: int | str | pd.Timestamp | None = None,
        window_size: int | None = None,
        differentiation: int | None = None,
        refit: bool | int = False,
        fixed_train_size: bool = True,
        gap: int = 0,
        skip_folds: int | list[int] | None = None,
        allow_incomplete_fold: bool = True,
        return_all_indexes: bool = False,
        verbose: bool = True
    ) -> None:

        self._validate_params(
            cv_name               = type(self).__name__,
            steps                 = steps,
            initial_train_size    = initial_train_size,
            window_size           = window_size,
            differentiation       = differentiation,
            refit                 = refit,
            fixed_train_size      = fixed_train_size,
            gap                   = gap,
            skip_folds            = skip_folds,
            allow_incomplete_fold = allow_incomplete_fold,
            return_all_indexes    = return_all_indexes,
            verbose               = verbose
        )

        self.steps                 = steps
        self.initial_train_size    = initial_train_size
        self.window_size           = window_size
        self.differentiation       = differentiation
        self.refit                 = refit
        self.fixed_train_size      = fixed_train_size
        self.gap                   = gap
        self.skip_folds            = skip_folds
        self.allow_incomplete_fold = allow_incomplete_fold
        self.return_all_indexes    = return_all_indexes
        self.verbose               = verbose

    def _validate_params(
        self,
        cv_name: str,
        steps: int | None = None,
        initial_train_size: int | str | pd.Timestamp | None = None,
        window_size: int | None = None,
        differentiation: int | None = None,
        refit: bool | int = False,
        fixed_train_size: bool = True,
        gap: int = 0,
        skip_folds: int | list[int] | None = None,
        allow_incomplete_fold: bool = True,
        return_all_indexes: bool = False,
        verbose: bool = True
    ) -> None: 
        """
        Validate all input parameters to ensure correctness.
        """

        if cv_name == "TimeSeriesFold":
            if not isinstance(steps, (int, np.integer)) or steps < 1:
                raise ValueError(
                    f"`steps` must be an integer greater than 0. Got {steps}."
                )
            if not isinstance(initial_train_size, (int, np.integer, str, pd.Timestamp, type(None))):
                raise ValueError(
                    f"`initial_train_size` must be an integer greater than 0, a date "
                    f"string, a pandas Timestamp, or None. Got {initial_train_size}."
                )
            if isinstance(initial_train_size, (int, np.integer)) and initial_train_size < 1:
                raise ValueError(
                    f"`initial_train_size` must be an integer greater than 0, "
                    f"a date string, a pandas Timestamp, or None. Got {initial_train_size}."
                )
            if not isinstance(refit, (bool, int, np.integer)):
                raise TypeError(
                    f"`refit` must be a boolean or an integer equal or greater than 0. "
                    f"Got {refit}."
                )
            if isinstance(refit, (int, np.integer)) and not isinstance(refit, bool) and refit < 0:
                raise TypeError(
                    f"`refit` must be a boolean or an integer equal or greater than 0. "
                    f"Got {refit}."
                )
            if not isinstance(fixed_train_size, bool):
                raise TypeError(
                    f"`fixed_train_size` must be a boolean: `True`, `False`. "
                    f"Got {fixed_train_size}."
                )
            if not isinstance(gap, (int, np.integer)) or gap < 0:
                raise ValueError(
                    f"`gap` must be an integer greater than or equal to 0. Got {gap}."
                )
            if skip_folds is not None:
                if not isinstance(skip_folds, (int, np.integer, list, type(None))):
                    raise TypeError(
                        f"`skip_folds` must be an integer greater than 0, a list of "
                        f"integers or `None`. Got {skip_folds}."
                    )
                if isinstance(skip_folds, (int, np.integer)) and skip_folds < 1:
                    raise ValueError(
                        f"`skip_folds` must be an integer greater than 0, a list of "
                        f"integers or `None`. Got {skip_folds}."
                    )
                if isinstance(skip_folds, list) and any([x < 1 for x in skip_folds]):
                    raise ValueError(
                        f"`skip_folds` list must contain integers greater than or "
                        f"equal to 1. The first fold is always needed to train the "
                        f"forecaster. Got {skip_folds}."
                    ) 
            if not isinstance(allow_incomplete_fold, bool):
                raise TypeError(
                    f"`allow_incomplete_fold` must be a boolean: `True`, `False`. "
                    f"Got {allow_incomplete_fold}."
                )
            
        if cv_name == "OneStepAheadFold":
            if not isinstance(initial_train_size, (int, np.integer, str, pd.Timestamp)):
                raise ValueError(
                    f"`initial_train_size` must be an integer greater than 0, a date "
                    f"string, or a pandas Timestamp. Got {initial_train_size}."
                )
            if isinstance(initial_train_size, (int, np.integer)) and initial_train_size < 1:
                raise ValueError(
                    f"`initial_train_size` must be an integer greater than 0, "
                    f"a date string, or a pandas Timestamp. Got {initial_train_size}."
                )
        
        if (
            not isinstance(window_size, (int, np.integer, pd.DateOffset, type(None)))
            or isinstance(window_size, (int, np.integer))
            and window_size < 1
        ):
            raise ValueError(
                f"`window_size` must be an integer greater than 0. Got {window_size}."
            )
        
        if not isinstance(return_all_indexes, bool):
            raise TypeError(
                f"`return_all_indexes` must be a boolean: `True`, `False`. "
                f"Got {return_all_indexes}."
            )
        if differentiation is not None:
            if not isinstance(differentiation, (int, np.integer)) or differentiation < 0:
                raise ValueError(
                    f"`differentiation` must be None or an integer greater than or "
                    f"equal to 0. Got {differentiation}."
                )
        if not isinstance(verbose, bool):
            raise TypeError(
                f"`verbose` must be a boolean: `True`, `False`. "
                f"Got {verbose}."
            )

    def _extract_index(
        self,
        X: pd.Series | pd.DataFrame | pd.Index | dict[str, pd.Series | pd.DataFrame]
    ) -> pd.Index:
        """
        Extracts and returns the index from the input data X.

        Parameters
        ----------
        X : pandas Series, pandas DataFrame, pandas Index, dict
            Time series data or index to split.

        Returns
        -------
        idx : pandas Index
            Index extracted from the input data.
        
        """

        if isinstance(X, (pd.Series, pd.DataFrame)):
            idx = X.index
        elif isinstance(X, dict):
            indexes_freq = set()
            not_valid_index = []
            min_index = []
            max_index = []
            for k, v in X.items():
                if v is None:
                    continue

                idx = v.index
                if isinstance(v.index, pd.DatetimeIndex):
                    indexes_freq.add(v.index.freqstr)
                elif isinstance(v.index, pd.RangeIndex):
                    indexes_freq.add(v.index.step)
                else:
                    not_valid_index.append(k)

                min_index.append(idx[0])
                max_index.append(idx[-1])

            if not_valid_index:
                raise TypeError(
                    f"If `X` is a dictionary, all series must have a Pandas "
                    f"RangeIndex or DatetimeIndex with the same step/frequency. "
                    f"Review series: {not_valid_index}"
                )

            if None in indexes_freq:
                raise ValueError(
                    "If `X` is a dictionary, all series must have a Pandas "
                    "RangeIndex or DatetimeIndex with the same step/frequency. "
                    "Found series with no frequency or step."
                )
            if not len(indexes_freq) == 1:
                raise ValueError(
                    f"If `X` is a dictionary, all series must have a Pandas "
                    f"RangeIndex or DatetimeIndex with the same step/frequency. "
                    f"Found frequencies: {sorted(indexes_freq)}"
                )
            
            if isinstance(idx, pd.DatetimeIndex):
                idx = pd.date_range(
                    start=min(min_index), end=max(max_index), freq=indexes_freq.pop()
                )
            else:
                idx = pd.RangeIndex(
                    start=min(min_index), stop=max(max_index) + 1, step=indexes_freq.pop()
                )
        else:
            idx = X
            
        return idx

    def set_params(
        self, 
        params: dict
    ) -> None:
        """
        Set the parameters of the Fold object. Before overwriting the current 
        parameters, the input parameters are validated to ensure correctness.

        Parameters
        ----------
        params : dict
            Dictionary with the parameters to set.
        
        Returns
        -------
        None
        
        """

        if not isinstance(params, dict):
            raise TypeError(
                f"`params` must be a dictionary. Got {type(params)}."
            )

        current_params = deepcopy(vars(self))
        unknown_params = set(params.keys()) - set(current_params.keys())
        if unknown_params:
            warnings.warn(
                f"Unknown parameters: {unknown_params}. They have been ignored.",
                IgnoredArgumentWarning
            )

        filtered_params = {k: v for k, v in params.items() if k in current_params}
        updated_params = {'cv_name': type(self).__name__, **current_params, **filtered_params}

        self._validate_params(**updated_params)
        for key, value in updated_params.items():
            setattr(self, key, value)


class OneStepAheadFold(BaseFold):
    """
    Class to split time series data into train and test folds for one-step-ahead
    forecasting.

    Parameters
    ----------
    initial_train_size : int, str, pandas Timestamp
        Number of observations used for initial training.

        - If an integer, the number of observations used for initial training.
        - If a date string or pandas Timestamp, it is the last date included in 
        the initial training set.
    window_size : int, default None
        Number of observations needed to generate the autoregressive predictors.
    differentiation : int, default None
        Number of observations to use for differentiation. This is used to extend the
        `last_window` as many observations as the differentiation order.
    return_all_indexes : bool, default False
        Whether to return all indexes or only the start and end indexes of each fold.
    verbose : bool, default True
        Whether to print information about generated folds.

    Attributes
    ----------
    initial_train_size : int
        Number of observations used for initial training.
    window_size : int
        Number of observations needed to generate the autoregressive predictors.
    differentiation : int 
        Number of observations to use for differentiation. This is used to extend the
        `last_window` as many observations as the differentiation order.
    return_all_indexes : bool
        Whether to return all indexes or only the start and end indexes of each fold.
    verbose : bool
        Whether to print information about generated folds.
    steps : Any
        This attribute is not used in this class. It is included for API consistency.
    fixed_train_size : Any
        This attribute is not used in this class. It is included for API consistency.
    gap : Any
        This attribute is not used in this class. It is included for API consistency.
    skip_folds : Any
        This attribute is not used in this class. It is included for API consistency.
    allow_incomplete_fold : Any
        This attribute is not used in this class. It is included for API consistency.
    refit : Any
        This attribute is not used in this class. It is included for API consistency.
    
    """

    def __init__(
        self,
        initial_train_size: int | str | pd.Timestamp,
        window_size: int | None = None,
        differentiation: int | None = None,
        return_all_indexes: bool = False,
        verbose: bool = True
    ) -> None:
        
        super().__init__(
            initial_train_size = initial_train_size,
            window_size        = window_size,
            differentiation    = differentiation,
            return_all_indexes = return_all_indexes,
            verbose            = verbose
        )

    def __repr__(
        self
    ) -> str:
        """
        Information displayed when printed.
        """

        info = (
            f"{'=' * len(type(self).__name__)} \n"
            f"{type(self).__name__} \n"
            f"{'=' * len(type(self).__name__)} \n"
            f"Initial train size = {self.initial_train_size},\n"
            f"Window size        = {self.window_size},\n"
            f"Differentiation    = {self.differentiation},\n"
            f"Return all indexes = {self.return_all_indexes},\n"
            f"Verbose            = {self.verbose}\n"
        )
            
        return info
    
    def _repr_html_(self) -> str:
        """
        HTML representation of the object.
        The "General Information" section is expanded by default.
        """

        style, unique_id = get_style_repr_html()
        content = f"""
        <div class="container-{unique_id}">
            <h2>{type(self).__name__}</h2>
            <details open>
                <summary>General Information</summary>
                <ul>
                    <li><strong>Initial train size:</strong> {self.initial_train_size}</li>
                    <li><strong>Window size:</strong> {self.window_size}</li>
                    <li><strong>Differentiation:</strong> {self.differentiation}</li>
                    <li><strong>Return all indexes:</strong> {self.return_all_indexes}</li>
                </ul>
            </details>
            <p>
                <a href="https://skforecast.org/{skforecast.__version__}/api/model_selection.html#skforecast.model_selection._split.OneStepAheadFold">&#128712 <strong>API Reference</strong></a>
                &nbsp;&nbsp;
                <a href="https://skforecast.org/{skforecast.__version__}/faq/parameters-search-backtesting-vs-one-step-ahead.html">&#128462 <strong>User Guide</strong></a>
            </p>
        </div>
        """

        return style + content
    
    def split(
        self,
        X: pd.Series | pd.DataFrame | pd.Index | dict[str, pd.Series | pd.DataFrame],
        as_pandas: bool = False,
        externally_fitted: Any = None
    ) -> list | pd.DataFrame:
        """
        Split the time series data into train and test folds.

        Parameters
        ----------
        X : pandas Series, DataFrame, Index, or dictionary
            Time series data or index to split.
        as_pandas : bool, default False
            If True, the folds are returned as a DataFrame. This is useful to visualize
            the folds in a more interpretable way.
        externally_fitted : Any
            This argument is not used in this class. It is included for API consistency.
        
        Returns
        -------
        fold : list, pandas DataFrame
            A list of lists containing the indices (position) of the fold. The list
            contains 2 lists with the following information:

            - [train_start, train_end]: list with the start and end positions of the
            training set.
            - [test_start, test_end]: list with the start and end positions of the test
            set. These are the observations used to evaluate the forecaster.
        
            It is important to note that the returned values are the positions of the
            observations and not the actual values of the index, so they can be used to
            slice the data directly using iloc.

            If `as_pandas` is `True`, the folds are returned as a DataFrame with the
            following columns: 'fold', 'train_start', 'train_end', 'test_start', 'test_end'.

            Following the python convention, the start index is inclusive and the end
            index is exclusive. This means that the last index is not included in the
            slice.
        
        """

        if not isinstance(X, (pd.Series, pd.DataFrame, pd.Index, dict)):
            raise TypeError(
                f"X must be a pandas Series, DataFrame, Index or a dictionary. "
                f"Got {type(X)}."
            )

        index = self._extract_index(X)

        self.initial_train_size = date_to_index_position(
                                      index        = index, 
                                      date_input   = self.initial_train_size, 
                                      method       = 'validation',
                                      date_literal = 'initial_train_size'
                                  )

        fold = [
            [0, self.initial_train_size - 1],
            [self.initial_train_size, len(X)],
            True
        ]

        if self.verbose:
            self._print_info(index=index, fold=fold)

        # NOTE: +1 to prevent iloc pandas from deleting the last observation
        if self.return_all_indexes:
            fold = [
                [range(fold[0][0], fold[0][1] + 1)],
                [range(fold[1][0], fold[1][1])],
                fold[2]
            ]
        else:
            fold = [
                [fold[0][0], fold[0][1] + 1],
                [fold[1][0], fold[1][1]],
                fold[2]
            ]

        if as_pandas:
            if not self.return_all_indexes:
                fold = pd.DataFrame(
                    data = [list(itertools.chain(*fold[:-1])) + [fold[-1]]],
                    columns = [
                        'train_start',
                        'train_end',
                        'test_start',
                        'test_end',
                        'fit_forecaster'
                    ],
                )
            else:
                fold = pd.DataFrame(
                    data = [fold],
                    columns = [
                        'train_index',
                        'test_index',
                        'fit_forecaster'
                    ],
                )
            fold.insert(0, 'fold', range(len(fold)))

        return fold

    def _print_info(
        self,
        index: pd.Index,
        fold: list[list[int]]
    ) -> None:
        """
        Print information about folds.

        Parameters
        ----------
        index : pandas Index
            Index of the time series data.
        fold : list
            A list of lists containing the indices (position) of the fold.

        Returns
        -------
        None

        """

        if self.differentiation is None:
            differentiation = 0
        else:
            differentiation = self.differentiation

        initial_train_size = self.initial_train_size - differentiation
        test_length = len(index) - (initial_train_size + differentiation)

        print("Information of folds")
        print("--------------------")
        print(
            f"Number of observations in train: {initial_train_size}"
        )
        if self.differentiation is not None:
            print(
                f"    First {differentiation} observation/s in training set "
                f"are used for differentiation"
            )
        print(
            f"Number of observations in test: {test_length}"
        )
        
        training_start = index[fold[0][0] + differentiation]
        training_end = index[fold[0][-1]]
        test_start  = index[fold[1][0]]
        test_end    = index[fold[1][-1] - 1]
        
        print(
            f"Training : {training_start} -- {training_end} (n={initial_train_size})"
        )
        print(
            f"Test     : {test_start} -- {test_end} (n={test_length})"
        )
        print("")


class TimeSeriesFold(BaseFold):
    """
    Class to split time series data into train and test folds. 
    When used within a backtesting or hyperparameter search, the arguments
    'initial_train_size', 'window_size' and 'differentiation' are not required
    as they are automatically set by the backtesting or hyperparameter search
    functions.

    Parameters
    ----------
    steps : int
        Number of observations used to be predicted in each fold. This is also commonly
        referred to as the forecast horizon or test size.
    initial_train_size : int, str, pandas Timestamp, default None
        Number of observations used for initial training. 
        
        - If `None` or 0, the initial forecaster is not trained in the first fold.
        - If an integer, the number of observations used for initial training.
        - If a date string or pandas Timestamp, it is the last date included in 
        the initial training set.
    window_size : int, default None
        Number of observations needed to generate the autoregressive predictors.
    differentiation : int, default None
        Number of observations to use for differentiation. This is used to extend the
        `last_window` as many observations as the differentiation order.
    refit : bool, int, default False
        Whether to refit the forecaster in each fold.

        - If `True`, the forecaster is refitted in each fold.
        - If `False`, the forecaster is trained only in the first fold.
        - If an integer, the forecaster is trained in the first fold and then refitted
          every `refit` folds.
    fixed_train_size : bool, default True
        Whether the training size is fixed or increases in each fold.
    gap : int, default 0
        Number of observations between the end of the training set and the start of the
        test set.
    skip_folds : int, list, default None
        Number of folds to skip.

        - If an integer, every 'skip_folds'-th is returned.
        - If a list, the indexes of the folds to skip.

        For example, if `skip_folds=3` and there are 10 folds, the returned folds are
        0, 3, 6, and 9. If `skip_folds=[1, 2, 3]`, the returned folds are 0, 4, 5, 6, 7,
        8, and 9.
    allow_incomplete_fold : bool, default True
        Whether to allow the last fold to include fewer observations than `steps`.
        If `False`, the last fold is excluded if it is incomplete.
    return_all_indexes : bool, default False
        Whether to return all indexes or only the start and end indexes of each fold.
    verbose : bool, default True
        Whether to print information about generated folds.

    Attributes
    ----------
    steps : int
        Number of observations used to be predicted in each fold. This is also commonly
        referred to as the forecast horizon or test size.
    initial_train_size : int
        Number of observations used for initial training. If `None` or 0, the initial
        forecaster is not trained in the first fold.
    window_size : int
        Number of observations needed to generate the autoregressive predictors.
    differentiation : int
        Number of observations to use for differentiation. This is used to extend the
        `last_window` as many observations as the differentiation order.
    refit : bool, int
        Whether to refit the forecaster in each fold.
    fixed_train_size : bool
        Whether the training size is fixed or increases in each fold.
    gap : int
        Number of observations between the end of the training set and the start of the
        test set.
    skip_folds : int, list
        Number of folds to skip.
    allow_incomplete_fold : bool
        Whether to allow the last fold to include fewer observations than `steps`.
    return_all_indexes : bool
        Whether to return all indexes or only the start and end indexes of each fold.
    verbose : bool
        Whether to print information about generated folds.

    Notes
    -----
    Returned values are the positions of the observations and not the actual values of
    the index, so they can be used to slice the data directly using iloc. For example,
    if the input series is `X = [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]`, the 
    `initial_train_size = 3`, `window_size = 2`, `steps = 4`, and `gap = 1`,
    the output of the first fold will: [[0, 3], [1, 3], [3, 8], [4, 8], True].

    The first list `[0, 3]` indicates that the training set goes from the first to the
    third observation. The second list `[1, 3]` indicates that the last window seen by
    the forecaster during training goes from the second to the third observation. The
    third list `[3, 8]` indicates that the test set goes from the fourth to the eighth
    observation. The fourth list `[4, 8]` indicates that the test set including the gap
    goes from the fifth to the eighth observation. The boolean `False` indicates that the
    forecaster should not be trained in this fold.

    Following the python convention, the start index is inclusive and the end index is
    exclusive. This means that the last index is not included in the slice.

    """

    def __init__(
        self,
        steps: int,
        initial_train_size: int | str | pd.Timestamp | None = None,
        window_size: int | None = None,
        differentiation: int | None = None,
        refit: bool | int = False,
        fixed_train_size: bool = True,
        gap: int = 0,
        skip_folds: int | list[int] | None = None,
        allow_incomplete_fold: bool = True,
        return_all_indexes: bool = False,
        verbose: bool = True
    ) -> None:
        
        super().__init__(
            steps                 = steps,
            initial_train_size    = initial_train_size,
            window_size           = window_size,
            differentiation       = differentiation,
            refit                 = refit,
            fixed_train_size      = fixed_train_size,
            gap                   = gap,
            skip_folds            = skip_folds,
            allow_incomplete_fold = allow_incomplete_fold,
            return_all_indexes    = return_all_indexes,
            verbose               = verbose
        )

    def __repr__(
        self
    ) -> str:
        """
        Information displayed when printed.
        """

        info = (
            f"{'=' * len(type(self).__name__)} \n"
            f"{type(self).__name__} \n"
            f"{'=' * len(type(self).__name__)} \n"
            f"Initial train size    = {self.initial_train_size},\n"
            f"Steps                 = {self.steps},\n"
            f"Window size           = {self.window_size},\n"
            f"Differentiation       = {self.differentiation},\n"
            f"Refit                 = {self.refit},\n"
            f"Fixed train size      = {self.fixed_train_size},\n"
            f"Gap                   = {self.gap},\n"
            f"Skip folds            = {self.skip_folds},\n"
            f"Allow incomplete fold = {self.allow_incomplete_fold},\n"
            f"Return all indexes    = {self.return_all_indexes},\n"
            f"Verbose               = {self.verbose}\n"
        )

        return info
    
    def _repr_html_(self) -> str:
        """
        HTML representation of the object.
        The "General Information" section is expanded by default.
        """

        style, unique_id = get_style_repr_html()
        content = f"""
        <div class="container-{unique_id}">
            <h2>{type(self).__name__}</h2>
            <details open>
                <summary>General Information</summary>
                <ul>
                    <li><strong>Initial train size:</strong> {self.initial_train_size}</li>
                    <li><strong>Steps:</strong> {self.steps}</li>
                    <li><strong>Window size:</strong> {self.window_size}</li>
                    <li><strong>Differentiation:</strong> {self.differentiation}</li>
                    <li><strong>Refit:</strong> {self.refit}</li>
                    <li><strong>Fixed train size:</strong> {self.fixed_train_size}</li>
                    <li><strong>Gap:</strong> {self.gap}</li>
                    <li><strong>Skip folds:</strong> {self.skip_folds}</li>
                    <li><strong>Allow incomplete fold:</strong> {self.allow_incomplete_fold}</li>
                    <li><strong>Return all indexes:</strong> {self.return_all_indexes}</li>
                </ul>
            </details>
            <p>
                <a href="https://skforecast.org/{skforecast.__version__}/api/model_selection.html#skforecast.model_selection._split.TimeSeriesFold">&#128712 <strong>API Reference</strong></a>
                &nbsp;&nbsp;
                <a href="https://skforecast.org/{skforecast.__version__}/user_guides/backtesting.html#timeseriesfold">&#128462 <strong>User Guide</strong></a>
            </p>
        </div>
        """
        
        return style + content

    def split(
        self,
        X: pd.Series | pd.DataFrame | pd.Index | dict[str, pd.Series | pd.DataFrame],
        as_pandas: bool = False
    ) -> list | pd.DataFrame:
        """
        Split the time series data into train and test folds.

        Parameters
        ----------
        X : pandas Series, pandas DataFrame, pandas Index, dict
            Time series data or index to split.
        as_pandas : bool, default False
            If True, the folds are returned as a DataFrame. This is useful to visualize
            the folds in a more interpretable way.

        Returns
        -------
        folds : list, pandas DataFrame
            A list of lists containing the indices (position) for each fold. Each list
            contains 4 lists and a boolean with the following information:

            - [train_start, train_end]: list with the start and end positions of the
            training set.
            - [last_window_start, last_window_end]: list with the start and end positions
            of the last window seen by the forecaster during training. The last window
            is used to generate the lags use as predictors. If `differentiation` is
            included, the interval is extended as many observations as the
            differentiation order. If the argument `window_size` is `None`, this list is
            empty.
            - [test_start, test_end]: list with the start and end positions of the test
            set. These are the observations used to evaluate the forecaster.
            - [test_start_with_gap, test_end_with_gap]: list with the start and end
            positions of the test set including the gap. The gap is the number of
            observations between the end of the training set and the start of the test
            set.
            - fit_forecaster: boolean indicating whether the forecaster should be fitted
            in this fold.

            It is important to note that the returned values are the positions of the
            observations and not the actual values of the index, so they can be used to
            slice the data directly using iloc.

            If `as_pandas` is `True`, the folds are returned as a DataFrame with the
            following columns: 'fold', 'train_start', 'train_end', 'last_window_start',
            'last_window_end', 'test_start', 'test_end', 'test_start_with_gap',
            'test_end_with_gap', 'fit_forecaster'.

            Following the python convention, the start index is inclusive and the end
            index is exclusive. This means that the last index is not included in the
            slice.

        """

        if not isinstance(X, (pd.Series, pd.DataFrame, pd.Index, dict)):
            raise TypeError(
                f"X must be a pandas Series, DataFrame, Index or a dictionary. "
                f"Got {type(X)}."
            )
        
        window_size_as_date_offset = isinstance(self.window_size, pd.tseries.offsets.DateOffset)
        if window_size_as_date_offset:
            # Calculate the window_size in steps. This is not a exact calculation
            # because the offset follows the calendar rules and the distance between
            # two dates may not be constant.
            first_valid_index = X.index[-1] - self.window_size
            try:
                window_size_idx_start = X.index.get_loc(first_valid_index)
                window_size_idx_end = X.index.get_loc(X.index[-1])
                self.window_size = window_size_idx_end - window_size_idx_start
            except KeyError:
                raise ValueError(
                    f"The length of `y` ({len(X)}), must be greater than or equal "
                    f"to the window size ({self.window_size}). This is because  "
                    f"the offset (forecaster.offset) is larger than the available "
                    f"data. Try to decrease the size of the offset (forecaster.offset), "
                    f"the number of `n_offsets` (forecaster.n_offsets) or increase the "
                    f"size of `y`."
                )
        
        if self.initial_train_size is None:
            if self.window_size is None:
                raise ValueError(
                    "To use split method when `initial_train_size` is None, "
                    "`window_size` must be an integer greater than 0. "
                    "Although no initial training is done and all data is used to "
                    "evaluate the model, the first `window_size` observations are "
                    "needed to create the initial predictors. Got `window_size` = None."
                )
            if self.refit:
                raise ValueError(
                    "`refit` is only allowed when `initial_train_size` is not `None`. "
                    "Set `refit` to `False` if you want to use `initial_train_size = None`."
                )
            externally_fitted = True
            self.initial_train_size = self.window_size  # Reset to None later
        else:
            if self.window_size is None:
                warnings.warn(
                    "Last window cannot be calculated because `window_size` is None.",
                    IgnoredArgumentWarning
                )
            externally_fitted = False

        index = self._extract_index(X)
        idx = range(len(index))
        folds = []
        i = 0
        last_fold_excluded = False

        self.initial_train_size = date_to_index_position(
                                      index        = index, 
                                      date_input   = self.initial_train_size, 
                                      method       = 'validation',
                                      date_literal = 'initial_train_size'
                                  )
        
        if window_size_as_date_offset:
            if self.initial_train_size is not None:
                if self.initial_train_size < self.window_size:
                    raise ValueError(
                        f"If `initial_train_size` is an integer, it must be greater than "
                        f"the `window_size` of the forecaster ({self.window_size}) "
                        f"and smaller than the length of the series ({len(X)}). If "
                        f"it is a date, it must be within this range of the index."
                    )

        if len(index) < self.initial_train_size + self.steps:
            raise ValueError(
                f"The time series must have at least `initial_train_size + steps` "
                f"observations. Got {len(index)} observations."
            )

        while self.initial_train_size + (i * self.steps) + self.gap < len(index):

            if self.refit:
                # If `fixed_train_size` the train size doesn't increase but moves by 
                # `steps` positions in each iteration. If `False`, the train size
                # increases by `steps` in each iteration.
                train_iloc_start = i * (self.steps) if self.fixed_train_size else 0
                train_iloc_end = self.initial_train_size + i * (self.steps)
                test_iloc_start = train_iloc_end
            else:
                # The train size doesn't increase and doesn't move.
                train_iloc_start = 0
                train_iloc_end = self.initial_train_size
                test_iloc_start = self.initial_train_size + i * (self.steps)
            
            if self.window_size is not None:
                last_window_iloc_start = test_iloc_start - self.window_size
            test_iloc_end = test_iloc_start + self.gap + self.steps
        
            partitions = [
                idx[train_iloc_start : train_iloc_end],
                idx[last_window_iloc_start : test_iloc_start] if self.window_size is not None else [],
                idx[test_iloc_start : test_iloc_end],
                idx[test_iloc_start + self.gap : test_iloc_end]
            ]
            folds.append(partitions)
            i += 1

        if not self.allow_incomplete_fold and len(folds[-1][3]) < self.steps:
            folds = folds[:-1]
            last_fold_excluded = True

        # Replace partitions inside folds with length 0 with `None`
        folds = [
            [partition if len(partition) > 0 else None for partition in fold] 
             for fold in folds
        ]

        # Create a flag to know whether to train the forecaster
        if self.refit == 0:
            self.refit = False
            
        if isinstance(self.refit, bool):
            fit_forecaster = [self.refit] * len(folds)
            fit_forecaster[0] = True
        else:
            fit_forecaster = [False] * len(folds)
            for i in range(0, len(fit_forecaster), self.refit): 
                fit_forecaster[i] = True
        
        for i in range(len(folds)): 
            folds[i].append(fit_forecaster[i])
            if fit_forecaster[i] is False:
                folds[i][0] = folds[i - 1][0]

        index_to_skip = []
        if self.skip_folds is not None:
            if isinstance(self.skip_folds, (int, np.integer)) and self.skip_folds > 0:
                index_to_keep = np.arange(0, len(folds), self.skip_folds)
                index_to_skip = np.setdiff1d(np.arange(0, len(folds)), index_to_keep, assume_unique=True)
                index_to_skip = [int(x) for x in index_to_skip]  # Required since numpy 2.0
            if isinstance(self.skip_folds, list):
                index_to_skip = [i for i in self.skip_folds if i < len(folds)]        
        
        if self.verbose:
            self._print_info(
                index              = index,
                folds              = folds,
                externally_fitted  = externally_fitted,
                last_fold_excluded = last_fold_excluded,
                index_to_skip      = index_to_skip
            )

        folds = [fold for i, fold in enumerate(folds) if i not in index_to_skip]
        if not self.return_all_indexes:
            # NOTE: +1 to prevent iloc pandas from deleting the last observation
            folds = [
                [[fold[0][0], fold[0][-1] + 1], 
                 [fold[1][0], fold[1][-1] + 1] if self.window_size is not None else [],
                 [fold[2][0], fold[2][-1] + 1],
                 [fold[3][0], fold[3][-1] + 1],
                 fold[4]] 
                for fold in folds
            ]

        if externally_fitted:
            self.initial_train_size = None
            folds[0][4] = False

        if as_pandas:
            if self.window_size is None:
                for fold in folds:
                    fold[1] = [None, None]

            if not self.return_all_indexes:
                folds = pd.DataFrame(
                    data = [list(itertools.chain(*fold[:-1])) + [fold[-1]] for fold in folds],
                    columns = [
                        'train_start',
                        'train_end',
                        'last_window_start',
                        'last_window_end',
                        'test_start',
                        'test_end',
                        'test_start_with_gap',
                        'test_end_with_gap',
                        'fit_forecaster'
                    ],
                )
            else:
                folds = pd.DataFrame(
                    data = folds,
                    columns = [
                        'train_index',
                        'last_window_index',
                        'test_index',
                        'test_index_with_gap',
                        'fit_forecaster'
                    ],
                )
            folds.insert(0, 'fold', range(len(folds)))

        return folds

    def _print_info(
        self,
        index: pd.Index,
        folds: list[list[int]],
        externally_fitted: bool,
        last_fold_excluded: bool,
        index_to_skip: list[int]
    ) -> None:
        """
        Print information about folds.

        Parameters
        ----------
        index : pandas Index
            Index of the time series data.
        folds : list
            A list of lists containing the indices (position) for each fold.
        externally_fitted : bool
            Whether an already trained forecaster is to be used.
        last_fold_excluded : bool
            Whether the last fold has been excluded because it was incomplete.
        index_to_skip : list
            Number of folds skipped.

        Returns
        -------
        None
        
        """

        print("Information of folds")
        print("--------------------")
        if externally_fitted:
            print(
                f"An already trained forecaster is to be used. Window size: "
                f"{self.window_size}"
            )
        else:
            if self.differentiation is None:
                print(
                    f"Number of observations used for initial training: "
                    f"{self.initial_train_size}"
                )
            else:
                print(
                    f"Number of observations used for initial training: "
                    f"{self.initial_train_size - self.differentiation}"
                )
                print(
                    f"    First {self.differentiation} observation/s in training sets "
                    f"are used for differentiation"
                )
        print(
            f"Number of observations used for backtesting: "
            f"{len(index) - self.initial_train_size}"
        )
        print(f"    Number of folds: {len(folds)}")
        print(
            f"    Number skipped folds: "
            f"{len(index_to_skip)} {index_to_skip if index_to_skip else ''}"
        )
        print(f"    Number of steps per fold: {self.steps}")
        print(
            f"    Number of steps to exclude between last observed data "
            f"(last window) and predictions (gap): {self.gap}"
        )
        if last_fold_excluded:
            print("    Last fold has been excluded because it was incomplete.")
        if len(folds[-1][3]) < self.steps:
            print(f"    Last fold only includes {len(folds[-1][3])} observations.")
        print("")

        if self.differentiation is None:
            differentiation = 0
        else:
            differentiation = self.differentiation
        
        for i, fold in enumerate(folds):
            is_fold_skipped   = i in index_to_skip
            has_training      = fold[-1] if i != 0 else True
            training_start    = (
                index[fold[0][0] + differentiation] if fold[0] is not None else None
            )
            training_end      = index[fold[0][-1]] if fold[0] is not None else None
            training_length   = (
                len(fold[0]) - differentiation if fold[0] is not None else 0
            )
            validation_start  = index[fold[3][0]]
            validation_end    = index[fold[3][-1]]
            validation_length = len(fold[3])

            print(f"Fold: {i}")
            if is_fold_skipped:
                print("    Fold skipped")
            elif not externally_fitted and has_training:
                print(
                    f"    Training:   {training_start} -- {training_end}  "
                    f"(n={training_length})"
                )
                print(
                    f"    Validation: {validation_start} -- {validation_end}  "
                    f"(n={validation_length})"
                )
            else:
                print("    Training:   No training in this fold")
                print(
                    f"    Validation: {validation_start} -- {validation_end}  "
                    f"(n={validation_length})"
                )

        print("")



# Unit test predict ForecasterRecursive
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.preprocessing import FunctionTransformer
from sklearn.compose import make_column_transformer
from sklearn.compose import make_column_selector
from sklearn.pipeline import make_pipeline
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import HistGradientBoostingRegressor
from lightgbm import LGBMRegressor

from skforecast.preprocessing import RollingFeatures
from skforecast.preprocessing import TimeSeriesDifferentiator
from skforecast.recursive import ForecasterRecursive

# Fixtures
from .fixtures_forecaster_recursive import y as y_categorical
from .fixtures_forecaster_recursive import exog as exog_categorical
from .fixtures_forecaster_recursive import exog_predict as exog_predict_categorical
from .fixtures_forecaster_recursive import data  # to test results when using differentiation


def test_predict_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict(steps=5)


def test_predict_output_when_regressor_is_LinearRegression():
    """
    Test predict output when using LinearRegression as regressor.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(50)))
    predictions = forecaster.predict(steps=5)
    expected = pd.Series(
                   data = np.array([50., 51., 52., 53., 54.]),
                   index = pd.RangeIndex(start=50, stop=55, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)

        
def test_predict_output_when_regressor_is_LinearRegression_with_exog():
    """
    Test predict output when using LinearRegression as regressor.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(50), name='y'), exog=pd.Series(np.arange(50, 150, 2), name='exog'))
    exog_pred = pd.Series(np.arange(100, 105), index=pd.RangeIndex(start=50, stop=55), name='exog')
    predictions = forecaster.predict(steps=5, exog=exog_pred)

    expected = pd.Series(
                   data = np.array([35.71428571428572, 34.38775510204082, 32.72886297376094,
                                    30.69012911286965, 30.258106741238777]),
                   index = pd.RangeIndex(start=50, stop=55, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_output_when_regressor_is_LinearRegression_with_transform_y():
    """
    Test predict output when using LinearRegression as regressor and StandardScaler.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4 ,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8 ,
                      -0.61, -0.88])
        )
    transformer_y = StandardScaler()
    forecaster = ForecasterRecursive(
                    regressor = LinearRegression(),
                    lags = 5,
                    transformer_y = transformer_y,
                )
    forecaster.fit(y=y)
    predictions = forecaster.predict(steps=5)

    expected = pd.Series(
                data = np.array([-0.1578203 , -0.18459942, -0.13711051, -0.01966358, -0.03228613]),
                index = pd.RangeIndex(start=20, stop=25, step=1),
                name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_output_when_regressor_is_LinearRegression_with_transform_y_and_transform_exog():
    """
    Test predict output when using LinearRegression as regressor, StandardScaler
    as transformer_y and transformer_exog as transformer_exog.
    """
    y = pd.Series(
            np.array([-0.59, 0.02, -0.9, 1.09, -3.61, 0.72, -0.11, -0.4])
        )
    exog = pd.DataFrame({
               'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4],
               'col_2': ['a', 'a', 'a', 'a', 'b', 'b', 'b', 'b']}
           )
    exog_predict = exog.copy()
    exog_predict.index = pd.RangeIndex(start=8, stop=16)

    transformer_y = StandardScaler()
    transformer_exog = ColumnTransformer(
                            [('scale', StandardScaler(), ['col_1']),
                             ('onehot', OneHotEncoder(), ['col_2'])],
                            remainder = 'passthrough',
                            verbose_feature_names_out = False
                       )
    forecaster = ForecasterRecursive(
                     regressor        = LinearRegression(),
                     lags             = 5,
                     transformer_y    = transformer_y,
                     transformer_exog = transformer_exog,
                 )
    forecaster.fit(y=y, exog=exog)
    predictions = forecaster.predict(steps=5, exog=exog_predict)

    expected = pd.Series(
                   data = np.array([0.50619336, -0.09630298,  0.05254973,  0.12281153,  0.00221741]),
                   index = pd.RangeIndex(start=8, stop=13, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_output_when_regressor_is_LinearRegression_and_weight_func():
    """
    Test predict output when using LinearRegression as regressor and custom_weights.
    """
    def custom_weights(index):
        """
        Return 1 for all elements in index
        """
        weights = np.ones_like(index)

        return weights

    forecaster = ForecasterRecursive(LinearRegression(), lags=3, weight_func=custom_weights)
    forecaster.fit(y=pd.Series(np.arange(50)))
    predictions = forecaster.predict(steps=5)

    expected = pd.Series(
                   data = np.array([50., 51., 52., 53., 54.]),
                   index = pd.RangeIndex(start=50, stop=55, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_output_when_categorical_features_native_implementation_HistGradientBoostingRegressor():
    """
    Test predict output when using HistGradientBoostingRegressor and categorical variables.
    """
    df_exog = pd.DataFrame({'exog_1': exog_categorical,
                            'exog_2': ['a', 'b', 'c', 'd', 'e']*10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J']*10)})
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursive(
                     regressor        = HistGradientBoostingRegressor(
                                            categorical_features = categorical_features,
                                            random_state         = 123
                                        ),
                     lags             = 5,
                     transformer_y    = None,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y_categorical, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=exog_predict)

    expected = pd.Series(
                   data = np.array([0.61187012, 0.42274801, 0.43214802, 0.4923281 , 
                                    0.53073262, 0.48004443, 0.56731689, 0.53956024, 
                                    0.47670124, 0.43896242]),
                   index = pd.RangeIndex(start=50, stop=60, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_output_when_categorical_features_native_implementation_LGBMRegressor():
    """
    Test predict output when using LGBMRegressor and categorical variables.
    """
    df_exog = pd.DataFrame({'exog_1': exog_categorical,
                            'exog_2': ['a', 'b', 'c', 'd', 'e']*10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J']*10)})
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursive(
                     regressor        = LGBMRegressor(random_state=123),
                     lags             = 5,
                     transformer_y    = None,
                     transformer_exog = transformer_exog,
                     fit_kwargs       = {'categorical_feature': categorical_features}
                 )
    forecaster.fit(y=y_categorical, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=exog_predict)

    expected = pd.Series(
                   data = np.array([0.5857033 , 0.3894503 , 0.45053399, 0.49686551, 
                                    0.45887492, 0.51481068, 0.5857033 , 0.3894503 , 
                                    0.45053399, 0.48818584]),
                   index = pd.RangeIndex(start=50, stop=60, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_output_when_categorical_features_native_implementation_LGBMRegressor_auto():
    """
    Test predict output when using LGBMRegressor and categorical variables with 
    categorical_features='auto'.
    """
    df_exog = pd.DataFrame({'exog_1': exog_categorical,
                            'exog_2': ['a', 'b', 'c', 'd', 'e']*10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J']*10)})
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    pipeline_categorical = make_pipeline(
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               FunctionTransformer(
                                   func=lambda x: x.astype('category'),
                                   feature_names_out= 'one-to-one'
                               )
                           )
    transformer_exog = make_column_transformer(
                            (
                                pipeline_categorical,
                                make_column_selector(dtype_exclude=np.number)
                            ),
                            remainder="passthrough",
                            verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursive(
                     regressor        = LGBMRegressor(random_state=123),
                     lags             = 5,
                     transformer_y    = None,
                     transformer_exog = transformer_exog,
                     fit_kwargs       = {'categorical_feature': 'auto'}
                 )
    forecaster.fit(y=y_categorical, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=exog_predict)

    expected = pd.Series(
                   data = np.array([0.5857033 , 0.3894503 , 0.45053399, 0.49686551, 
                                    0.45887492, 0.51481068, 0.5857033 , 0.3894503 , 
                                    0.45053399, 0.48818584]),
                   index = pd.RangeIndex(start=50, stop=60, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_output_when_regressor_is_LinearRegression_with_exog_and_differentiation_is_1():
    """
    Test predict output when using LinearRegression as regressor and differentiation=1.
    """

    # Data differentiated
    differentiator = TimeSeriesDifferentiator(order=1)
    data_diff = differentiator.fit_transform(data.to_numpy())
    data_diff = pd.Series(data_diff, index=data.index).dropna()
    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff = exog.iloc[1:]
    end_train = '2003-03-01 23:59:00'
    steps = len(data.loc[end_train:])

    forecaster_1 = ForecasterRecursive(regressor=LinearRegression(), lags=15)
    forecaster_1.fit(y=data_diff.loc[:end_train], exog=exog_diff.loc[:end_train])
    predictions_diff = forecaster_1.predict(steps=steps, exog=exog_diff.loc[end_train:])
    # Revert the differentiation
    last_value_train = data.loc[:end_train].iloc[[-1]]
    predictions_1 = pd.concat([last_value_train, predictions_diff]).cumsum()[1:]

    forecaster_2 = ForecasterRecursive(regressor=LinearRegression(), lags=15, differentiation=1)
    forecaster_2.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])
    predictions_2 = forecaster_2.predict(steps=steps, exog=exog.loc[end_train:])

    pd.testing.assert_series_equal(predictions_1.asfreq('MS'), predictions_2, check_names=False)


def test_predict_output_when_regressor_is_LinearRegression_with_exog_differentiation_is_1_and_transformer_y():
    """
    Test predict output when using LinearRegression as regressor and differentiation=1,
    and transformer_y is StandardScaler.
    """

    end_train = '2003-03-01 23:59:00'
    # Data scaled and differentiated
    scaler = StandardScaler()
    scaler.fit(data.loc[:end_train].to_numpy().reshape(-1, 1))
    data_scaled = scaler.transform(data.to_numpy().reshape(-1, 1))
    data_scaled = pd.Series(data_scaled.ravel(), index=data.index)
    data_scaled_diff = TimeSeriesDifferentiator(order=1).fit_transform(data_scaled.to_numpy())
    data_scaled_diff = pd.Series(data_scaled_diff, index=data.index).dropna()
    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff = exog.iloc[1:]
    steps = len(data.loc[end_train:])

    forecaster_1 = ForecasterRecursive(regressor=LinearRegression(), lags=15)
    forecaster_1.fit(y=data_scaled_diff.loc[:end_train], exog=exog_diff.loc[:end_train])
    predictions_diff = forecaster_1.predict(steps=steps, exog=exog_diff.loc[end_train:])
    # Revert the differentiation
    last_value_train = data_scaled.loc[:end_train].iloc[[-1]]
    predictions_1 = pd.concat([last_value_train, predictions_diff]).cumsum()[1:]
    # Revert the scaling
    predictions_1 = scaler.inverse_transform(predictions_1.to_numpy().reshape(-1, 1))
    predictions_1 = pd.Series(predictions_1.ravel(), index=data.loc[end_train:].index)

    forecaster_2 = ForecasterRecursive(regressor=LinearRegression(), lags=15, differentiation=1)
    forecaster_2.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])
    predictions_2 = forecaster_2.predict(steps=steps, exog=exog.loc[end_train:])

    pd.testing.assert_series_equal(predictions_1.asfreq('MS'), predictions_2, check_names=False)


def test_predict_output_when_regressor_is_LinearRegression_with_exog_and_differentiation_is_2():
    """
    Test predict output when using LinearRegression as regressor and differentiation=2.
    """

    # Data differentiated
    differentiator_1 = TimeSeriesDifferentiator(order=1)
    differentiator_2 = TimeSeriesDifferentiator(order=2)
    data_diff_1 = differentiator_1.fit_transform(data.to_numpy())
    data_diff_1 = pd.Series(data_diff_1, index=data.index).dropna()
    data_diff_2 = differentiator_2.fit_transform(data.to_numpy())
    data_diff_2 = pd.Series(data_diff_2, index=data.index).dropna()

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff_2 = exog.iloc[2:]
    end_train = '2003-03-01 23:59:00'
    steps = len(data.loc[end_train:])

    forecaster_1 = ForecasterRecursive(regressor=LinearRegression(), lags=15)
    forecaster_1.fit(y=data_diff_2.loc[:end_train], exog=exog_diff_2.loc[:end_train])
    predictions_diff_2 = forecaster_1.predict(steps=steps, exog=exog_diff_2.loc[end_train:])
    
    # Revert the differentiation
    last_value_train_diff = data_diff_1.loc[:end_train].iloc[[-1]]
    predictions_diff_1 = pd.concat([last_value_train_diff, predictions_diff_2]).cumsum()[1:]
    last_value_train = data.loc[:end_train].iloc[[-1]]
    predictions_1 = pd.concat([last_value_train, predictions_diff_1]).cumsum()[1:]

    forecaster_2 = ForecasterRecursive(regressor=LinearRegression(), lags=15, differentiation=2)
    forecaster_2.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])
    predictions_2 = forecaster_2.predict(steps=steps, exog=exog.loc[end_train:])

    pd.testing.assert_series_equal(predictions_1.asfreq('MS'), predictions_2, check_names=False)


@pytest.mark.parametrize("steps", 
                         [10, '2001-03-01', pd.to_datetime('2001-03-01')], 
                         ids=lambda steps: f'steps: {steps}')
def test_predict_output_when_window_features(steps):
    """
    Test output of predict when regressor is LGBMRegressor and window features.
    """
    y_datetime = y_categorical.copy()
    y_datetime.index = pd.date_range(start='2001-01-01', periods=len(y_datetime), freq='D')
    exog_datetime = exog_categorical.copy()
    exog_datetime.index = pd.date_range(start='2001-01-01', periods=len(exog_datetime), freq='D')
    exog_predict_datetime = exog_predict_categorical.copy()
    exog_predict_datetime.index = pd.date_range(start='2001-02-20', periods=len(exog_predict_datetime), freq='D')
    
    rolling = RollingFeatures(stats=['mean', 'sum'], window_sizes=[3, 5])
    forecaster = ForecasterRecursive(
        LGBMRegressor(verbose=-1, random_state=123), lags=3, window_features=rolling
    )
    forecaster.fit(y=y_datetime, exog=exog_datetime)
    predictions = forecaster.predict(steps=steps, exog=exog_predict_datetime)

    expected = pd.Series(
                   data = np.array([0.5326654111553376, 0.5050280233102159, 
                                    0.5050280233102159, 0.5050280233102159, 
                                    0.5326654111553376, 0.5326654111553376, 
                                    0.5326654111553376, 0.5050280233102159, 
                                    0.5050280233102159, 0.5326654111553376]),
                   index = pd.date_range(start='2001-02-20', periods=10, freq='D'),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)



# Unit test date_to_index_position
# ==============================================================================
import re
import pytest
import pandas as pd
from skforecast.utils import date_to_index_position


def test_ValueError_date_to_index_position_when_method_not_valid():
    """
    Test ValueError is raised when `method` is not 'prediction' or 'validation'.
    """
    index = pd.date_range(start='1990-01-01', periods=3, freq='D')
    
    err_msg = re.escape("`method` must be 'prediction' or 'validation'.")
    with pytest.raises(ValueError, match=err_msg):
        date_to_index_position(index, date_input='1990-01-10', method='not_valid')


def test_TypeError_date_to_index_position_when_index_is_not_DatetimeIndex():
    """
    Test TypeError is raised when `date_input` is a date but the index is not 
    a DatetimeIndex.
    """
    index = pd.RangeIndex(start=0, stop=3, step=1)
    
    err_msg = re.escape(
        "Index must be a pandas DatetimeIndex when `steps` is not an integer. "
        "Check input series or last window."
    )
    with pytest.raises(TypeError, match=err_msg):
        date_to_index_position(index, date_input='1990-01-10')


def test_ValueError_date_to_index_position_when_date_is_before_last_index():
    """
    Test ValueError is raised when the provided date is earlier than or equal 
    to the last date in index.
    """
    index = pd.date_range(start='1990-01-01', periods=3, freq='D')
    
    err_msg = re.escape(
        "If `steps` is a date, it must be greater than the last date "
        "in the index."
    )
    with pytest.raises(ValueError, match=err_msg):
        date_to_index_position(index, date_input='1990-01-02')


def test_TypeError_date_to_index_position_when_date_input_is_not_int_str_or_Timestamp():
    """
    Test TypeError is raised when `date_input` is not a int, str or pd.Timestamp.
    """
    index = pd.date_range(start='1990-01-01', periods=3, freq='D')
    date_input = 2.5
    date_literal = 'initial_train_size'
    
    err_msg = re.escape(
        "`initial_train_size` must be an integer, string, or pandas Timestamp."
    )
    with pytest.raises(TypeError, match=err_msg):
        date_to_index_position(
            index=index, date_input=date_input, date_literal=date_literal
        )


@pytest.mark.parametrize("date_input", 
                         ['1990-01-07', pd.Timestamp('1990-01-07'), 4], 
                         ids = lambda date_input: f'date_input: {type(date_input)}')
def test_output_date_to_index_position_with_different_date_input_types(date_input):
    """
    Test values returned by date_to_index_position with different date_input types.
    """
    index = pd.date_range(start='1990-01-01', periods=3, freq='D')
    results = date_to_index_position(index=index, date_input=date_input)

    expected = 4
    
    assert results == expected


def test_output_date_to_index_position_when_date_input_is_string_date_with_kwargs_pd_to_datetime():
    """
    Test values returned by date_to_index_position when `date_input` is a string 
    date and `kwargs_pd_to_datetime` are passed.
    """
    index = pd.date_range(start='1990-01-01', periods=3, freq='D')
    results = date_to_index_position(
        index=index, date_input='1990-07-01', kwargs_pd_to_datetime={'format': '%Y-%d-%m'}
    )
    
    expected = 4
    
    assert results == expected


def test_ValueError_date_to_index_position_when_date_is_out_of_range_and_method_is_validation():
    """
    Test ValueError is raised when date_input is out of the index range
    and method is 'validation'.
    """
    index = pd.date_range(start='1990-01-01', periods=3, freq='D')
    
    err_msg = re.escape(
        "If `initial_train_size` is a date, it must be greater than "
        "the first date in the index and less than the last date."
    )
    with pytest.raises(ValueError, match=err_msg):
        date_to_index_position(
            index        = index,
            date_input   = '1990-01-10',
            method       = 'validation',
            date_literal ='initial_train_size'
        )


def test_output_date_to_index_position_when_date_in_range_and_method_is_validation():
    """
    Test correct position is returned when date_input is within the index range
    and method is 'validation'.
    """
    index = pd.date_range(start='1990-01-01', periods=5, freq='D')
    results = date_to_index_position(
                  index        = index,
                  date_input   = '1990-01-03',
                  method       = 'validation',
                  date_literal ='initial_train_size'
              )

    expected = 3  # iloc position within the range

    assert results == expected


def test_output_date_to_index_position_when_date_is_first_date_and_method_is_validation():
    """
    Test it returns the correct position when date_input is exactly the first date
    in the index and method is 'validation'.
    """
    index = pd.date_range(start='1990-01-01', periods=5, freq='D')
    results = date_to_index_position(
        index=index, date_input='1990-01-01', method='validation', date_literal ='initial_train_size'
    )

    assert results == 1  # iloc position within the range
